#include <iostream>
#include <fluxions/Simple3DEngine.hpp>

using namespace std;


namespace Fluxions
{
	Simple3DEngine::Simple3DEngine()
	{
	}


	Simple3DEngine::~Simple3DEngine()
	{
	}


	void Simple3DEngine::Update()
	{
		SceneGraph.UpdateMatrices();
	}


	ShaderProgram *Simple3DEngine::GetShaderProgram(const std::string &name)
	{
		std::map<std::string, ShaderProgram>::iterator it = Shaders.find(name);
		if (it == Shaders.end())
			return NULL;
		else
			return &(it->second);
	}


	void Simple3DEngine::SetGlobalShaderOverride(const std::string &name)
	{
		globalShaderOverride = name;
	}


	void Simple3DEngine::SetGlobalShaderUniforms(UniformLibrary &uniforms)
	{
		std::map<std::string, ShaderProgram>::iterator it;
		for (it = Shaders.begin(); it != Shaders.end(); it++)
		{
			it->second.Use();
			it->second.SetUniforms(uniforms);
		}
		glUseProgram(0);
	}


	void Simple3DEngine::Render(bool fixedFunction)
	{
		std::list<SceneGraphNode*> nodes;
		std::list<SceneGraphNode*>::iterator it;
		SceneGraph.Flatten(nodes);

		for (it = nodes.begin(); it != nodes.end(); it++)
		{
			if (!(*it) || (*it)->StaticMeshName == "")
				continue;
			if (StaticMeshes.find((*it)->StaticMeshName) != StaticMeshes.end())
			{
				ShaderProgram *program = nullptr;
				if (fixedFunction != true)
				{
					if (globalShaderOverride == "")
						program = GetShaderProgram((*it)->ShaderName);
					else
						program = GetShaderProgram(globalShaderOverride);
					if (program != NULL)
					{
						program->Use();
						program->SetUniforms((*it)->Uniforms);
					}
					StaticMeshes[(*it)->StaticMeshName].SetShaderProgram(program);
				}
				else
				{
					glPushMatrix();
					glMultMatrixd((*it)->WorldMatrix.m);
				}
				StaticMeshes[(*it)->StaticMeshName].Render(-1, fixedFunction);
				if (fixedFunction == true)
				{
					glPopMatrix();
				}
				else
				{
					if (program != NULL)
					{
						glUseProgram(0);
					}
				}
			}
		}
	}


	void Simple3DEngine::CompileMeshes()
	{
		std::map<std::string, StaticModel>::iterator m;
		for (m = StaticMeshes.begin(); m != StaticMeshes.end(); m++)
		{
			m->second.Compile();
		}
	}


	void Simple3DEngine::ClearState()
	{
		ClearShaderState();
		ClearTextureState();
		ClearFrameBufferState();
		ClearVertexBufferState();

		glBindTransformFeedback(GL_TRANSFORM_FEEDBACK, 0);
	}


	void Simple3DEngine::ClearShaderState()
	{
		glUseProgram(0);
	}


	void Simple3DEngine::ClearTextureState()
	{
		int maxTextureUnits = 0;
		glGetIntegerv(GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS, &maxTextureUnits);
		if (maxTextureUnits>=32)
			maxTextureUnits = 32;

		for (int i = 0; i < maxTextureUnits; i++)
		{
			glActiveTexture(GL_TEXTURE0+i);
			glBindSampler(i, 0);
			glBindTexture(GL_TEXTURE_1D, 0);
			glBindTexture(GL_TEXTURE_2D, 0);
			glBindTexture(GL_TEXTURE_1D_ARRAY, 0);
			glBindTexture(GL_TEXTURE_2D_ARRAY, 0);
			glBindTexture(GL_TEXTURE_3D, 0);
			glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
			glBindTexture(GL_TEXTURE_RECTANGLE, 0);
			//glDisable(GL_TEXTURE_2D);
			//glDisable(GL_TEXTURE_1D);
			//glDisable(GL_TEXTURE_3D);
			//glDisable(GL_TEXTURE_2D_ARRAY);
			//glDisable(GL_TEXTURE_1D_ARRAY);
			//glDisable(GL_TEXTURE_CUBE_MAP);
			//glDisable(GL_TEXTURE_RECTANGLE);
		}
		glActiveTexture(GL_TEXTURE0);
	}


	void Simple3DEngine::ClearFrameBufferState()
	{
	}


	void Simple3DEngine::ClearVertexBufferState()
	{
		int maxVertexAttribs = 0;
		glGetIntegerv(GL_MAX_VERTEX_ATTRIBS, &maxVertexAttribs);
		glBindVertexArray(0);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
		for (int i = 0; i < maxVertexAttribs; i++)
		{
			glDisableVertexAttribArray(i);
		}
	}
}
