#include <fluxions/SimpleSceneGraph.hpp>

namespace Fluxions
{
	SceneGraphNode::SceneGraphNode()
	{
	}


	SceneGraphNode::~SceneGraphNode()
	{
		Clear();
	}


	SceneGraphNode &SceneGraphNode::operator[](const std::string &childName)
	{
		return children[childName];
	}


	SceneGraphNode &SceneGraphNode::GetChild(const std::string &childName)
	{
		return children[childName];
	}


	void SceneGraphNode::Clear()
	{
		SceneGraphNode::ChildIterator it;
		children.clear();
	}


	// Returns a list of visible nodes from this node and all child nodes
	void SceneGraphNode::Flatten(std::list<SceneGraphNode*> &nodeList)
	{
		SceneGraphNode::ChildIterator it;

		if (IsVisible())
		{
			nodeList.push_back(this);
		}

		for (it = children.begin(); it != children.end(); it++)
		{
			it->second.Flatten(nodeList);
		}
	}


	void SceneGraphNode::UpdateMatrices()
	{
		WorldMatrix *= LocalMatrix;
		// Update uniform
		Uniforms["gl_ModelViewMatrix"] = (ftl::Matrix4f)WorldMatrix;
		ChildIterator it;

		for (it = children.begin(); it != children.end(); it++)
		{
			it->second.WorldMatrix = WorldMatrix;
			it->second.UpdateMatrices();
		}
	}


	void SceneGraphNode::Update(double deltaTime)
	{
		ChildIterator it;

		OnUpdate(deltaTime);

		for (it = children.begin(); it != children.end(); it++)
		{
			it->second.Update(deltaTime);
		}
	}


	void SceneGraphNode::OnStart()
	{
	}


	void SceneGraphNode::OnUpdate(double deltaTime)
	{
	}
}
