#ifndef SIMPLE_3D_ENGINE_HPP
#define SIMPLE_3D_ENGINE_HPP

#include <fluxions/opengl.hpp>
#include <fluxions/SimpleSceneGraph.hpp>
#include <fluxions/staticmodel.hpp>
#include <fluxions/shaderprogram.hpp>
#include <fluxions/texture.hpp>
#include "fluxions_deferred_renderer.hpp"

namespace Fluxions
{

	class Simple3DEngine
	{
	public:
		Simple3DEngine();
		~Simple3DEngine();

		void Render(bool fixedFunction = false);
		void Update();
		void CompileMeshes();
		void SetGlobalShaderOverride(const std::string &name);
		void SetGlobalShaderUniforms(UniformLibrary &uniforms);

		void ClearState();
		void ClearShaderState();
		void ClearTextureState();
		void ClearFrameBufferState();
		void ClearVertexBufferState();

		SceneGraphNode SceneGraph;
		std::map<std::string, StaticModel> StaticMeshes;
		//std::map<std::string, GLuint> Textures;
		std::map<std::string, ShaderProgram> Shaders;
		std::map<std::string, Texture> Textures;

		// Deferred Rendering!
		void CreateDeferredRenderer(int imageWidth, int imageHeight, GLenum depthStencilFormat, int numColorAttachments, GLenum colorAttachmentFormat);
		void DeleteDeferredRenderer();
		void SetDeferredRenderingStage(DeferredRenderingStageTypeEnum stageType);
		void BindDeferredRenderingTextures(GLenum startTexture);
	private:
		ShaderProgram *GetShaderProgram(const std::string &name);
		std::string globalShaderOverride;
	};
}

#endif
