#ifndef SIMPLE_SCENE_GRAPH_HPP
#define SIMPLE_SCENE_GRAPH_HPP

#include <memory>
#include <string>
#include <vector>
#include <list>
#include <map>
#include <algorithm>

#include <fluxions/shaderprogram.hpp>
#include <fluxions/ftl.hpp>

namespace Fluxions
{
	class SceneGraphNode
		: public std::enable_shared_from_this<SceneGraphNode>
	{
	private:
		std::map<std::string, SceneGraphNode> children;
		bool visible;
	public:
		typedef std::map<std::string, SceneGraphNode>::iterator ChildIterator;

		std::string StaticMeshName;
		std::string ShaderName;
		UniformLibrary Uniforms;

		ftl::Matrix4d LocalMatrix;
		ftl::Matrix4d WorldMatrix;

		SceneGraphNode();
		virtual ~SceneGraphNode();

		inline void RemoveNode(const std::string &childName) { children.erase(childName); }
		void Clear();
		inline bool IsChild(const std::string &childName) { return children.find(childName) != children.end(); }
		inline bool IsLeaf() const { return children.empty(); }
		SceneGraphNode &operator[] (const std::string &childName);
		SceneGraphNode &GetChild(const std::string &childName);
		bool IsVisible() { return true; }
		void Hide() { visible = false; }
		void Show() { visible = true; }

		void Flatten(std::list<SceneGraphNode*> &nodeList);

		void UpdateMatrices();
		void Update(double elapsedTime);

		void OnStart();
		void OnUpdate(double elapsedTime);
	};
}

#endif
