#ifndef FLUXIONS_DEFERRED_RENDERER_HPP
#define FLUXIONS_DEFERRED_RENDERER_HPP

#include <vector>
#include <fluxions/ftl.hpp>
#include "opengl.hpp"

namespace Fluxions
{
	using namespace std;

	class Simple3DEngine;


	enum DeferredRenderingStageTypeEnum
	{
		DRST_NONE,
		DRST_GBUFFER_PASS,
		DRST_LIGHT_PASS,
		DRST_COMBINE_PASS,
		DRST_FINAL_PASS
	};


	class SimpleDeferredRenderer
	{
	public:
		SimpleDeferredRenderer();
		~SimpleDeferredRenderer();

		void Create(int imageWidth_, int imageHeight_, GLenum depthStencilFormat, int numColorAttachments, GLenum colorAttachmentFormat);
		bool CheckForCompleteness();
		void Delete();
		void Blit(int x = 0, int y = 0, int fbo = 0);
		void BlitGBuffer(int x = 0, int y = 0, int which = 0);
		void BlitAccumBuffer(int x = 0, int y = 0, int which = -1);
		void SetRenderingStage(DeferredRenderingStageTypeEnum stageType);
		void BindTextures(GLenum startTexture, bool unbind=false);

		void SetupGBufferPass(const ftl::Matrix4f &projectionMatrix, const ftl::Matrix4f &viewMatrix);
		void RenderGBufferPass(Simple3DEngine &scenegraph);
		void EndGBufferPass();

		void ClearAccumBuffer();
		void SetupDeferredLightingPass(ftl::Vector4f &lightPosition, ftl::Matrix4f &shadowMatrix, Simple3DEngine &scenegraph);
		void RenderDeferredLightingPass();
		void EndDeferredLightingPass();
	private:
		GLuint framebuffers[3];	// one is the gbuffer, the second and third is the accumulation buffer
		GLuint depthStencilTexture;
		vector<GLuint> colorAttachmentTextures;
		GLuint accumBufferTextures[2];
		int currentAccumBufferTexture;
		int imageWidth;
		int imageHeight;
		int textureWidth;
		int textureHeight;

		ftl::Matrix4f projectionMatrix;
		ftl::Matrix4f viewMatrix;

		// Deferred Lighting Pass
		ftl::Vector3f lightPosition;
		int vloc;
		int tloc;

		int NextPowerOf2(int x);
		GLuint CreateTexture(GLenum target, GLenum internalformat, GLsizei width, GLsizei height, GLenum format=GL_RGBA, GLenum type=GL_UNSIGNED_BYTE, const GLvoid *pixels=0, GLint baseLevel=0, GLint maxLevel=0);
	};
}

#endif
