#ifndef FLUXIONS_LENS_CAMERA_HPP
#define FLUXIONS_LENS_CAMERA_HPP


#include <fluxions/ftl.hpp>
#include <fluxions/opengl.hpp>


#define FLUXIONS_DIRECTION_LEFT		GL_TEXTURE_CUBE_MAP_NEGATIVE_X
#define FLUXIONS_DIRECTION_RIGHT	GL_TEXTURE_CUBE_MAP_POSITIVE_X
#define FLUXIONS_DIRECTION_TOP		GL_TEXTURE_CUBE_MAP_POSITIVE_Y
#define FLUXIONS_DIRECTION_BOTTOM	GL_TEXTURE_CUBE_MAP_NEGATIVE_Y
#define FLUXIONS_DIRECTION_FORWARD	GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
#define FLUXIONS_DIRECTION_BACKWARD	GL_TEXTURE_CUBE_MAP_POSITIVE_Z
#define FLUXIONS_CLEAR_NONE		0
#define FLUXIONS_CLEAR_COLOR	GL_COLOR_BUFFER_BIT
#define FLUXIONS_CLEAR_DEPTH	GL_DEPTH_BUFFER_BIT
#define FLUXIONS_CLEAR_STENCIL	GL_STENCIL_BUFFER_BIT
#define FLUXIONS_CLEAR_DEPTH_STENCIL	(GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT)
#define FLUXIONS_CLEAR_COLOR_DEPTH_STENCIL	(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT)

namespace Fluxions
{


class LensCamera
{
public:
	LensCamera();
	~LensCamera();

	void Init(int width, int height, float znear, float zfar, bool useDepthBuffer = true, GLenum depthBufferFunc = GL_LESS);
	void QuickSet4x5(float focalLength);		// 4x5 Large Format Camera
	void QuickSetAPS(float focalLength);		// Nikon 24mm Photo Camera (4:3) ratio
	void QuickSet35mm(float focalLength);		// 35mm Photo Camera (4:3) ratio
	void QuickSetSuper65(float focalLength);	// 70mm Video Camera (2.76 Aspect Ratio)
	void SetViewport(int screenWidth_, int screenHeight_);

	ftl::Matrix4f &GetViewProjectionMatrix() { return viewProjectionMatrix; }
	ftl::Matrix4f &GetProjectionLensMatrix() { return projectionMatrix * lensMatrix; }
	ftl::Matrix4f &GetWorldMatrix() { return worldMatrix * orientationMatrix; }
	ftl::Vector4f &GetWorldPosition() { return worldPosition; }

	void SetWorldPosition(GLfloat x, GLfloat y, GLfloat z);
	void SetPositionOrientation(GLfloat x, GLfloat y, GLfloat z, GLfloat theta, GLfloat phi);
	void SetLookAt(float x, float y, float z, float upx, float upy, float upz);
	void SetLookAt(float theta, float phi);
	void MoveRotate(float rx, float ry, float rz, float tx, float ty, float tz);

	float GetAspectRatio() const { return imageWidth / imageHeight; }
	float GetFocalLength() const { return focalLength; }
	float GetImageWidth() const { return imageWidth; }
	float GetImageHeight() const { return imageHeight; }
	int GetScreenWidth() const { return screenWidth; }
	int GetScreenHeight() const { return screenHeight; }
	void SetClearColor(float r, float g, float b, float a = 1.0f);
	void SetClearColor(const float *rgba);

	void RecalculateCamera();
	void SetViewport();
	void BeginFrame(GLbitfield clearBufferMask = FLUXIONS_CLEAR_COLOR_DEPTH_STENCIL,
					GLenum direction = FLUXIONS_DIRECTION_FORWARD);
	void EndFrame();
private:
	ftl::Matrix4f viewProjectionMatrix;
	ftl::Matrix4f projectionMatrix;
	ftl::Matrix4f lensMatrix;
	ftl::Matrix4f worldMatrix;
	ftl::Matrix4f orientationMatrix;
	ftl::Vector4f worldPosition;

	GL_Viewport savedViewport;

	float focalLength;  // focal length of lens (in mm)
	float imageWidth;	// sensor Width (in mm)
	float imageHeight;	// sensor Height (in mm)
	float imageAspectRatio;	// aspect ratio of sensor [computed by RecalculateCamera()]
	float coc;			// circle of confusion (in mm) - for depth of field equations

	int screenWidth;
	int screenHeight;

	int screenImageX;		// computed by RecalculateCamera()
	int screenImageY;		// computed by RecalculateCamera()
	int screenImageWidth;	// computed by RecalculateCamera()
	int screenImageHeight;	// computed by RecalculateCamera()

	float znear;
	float zfar;
	bool useDepthBuffer;
	GLenum depthBufferFunc;
	ftl::Vector4f clearColor;
};


}


#endif
