/*
Fluxions 3D Graphics Engine
Copyright (C) 2002  Jonathan Metzgar

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef FLUXIONS_PROCEDURAL_TEXTURE_H
#define FLUXIONS_PROCEDURAL_TEXTURE_H

#define STEP(a, x)		((float)((x) >= (a)))
#define PULSE(a, b, x)	(STEP((a),(x)) - STEP((b),(x)))
#define CLAMP(x, a, b)	((x) < (a) ? (a) : (x) > (b) ? (b) : (x))
#ifndef MIN
#define MIN(a, b)		((a) < (b) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(a, b)		((a) < (b) ? (b) : (a))
#endif
#define ABS(x)			((x) < 0 ? -(x) : (x))
#define FLOOR(x)		((int)(x) - ((x) < 0 && (x) != (int)(x)))
#define CEIL(x)			((int)(x) + ((x) > 0 && (x) != (int)(x)))
#define BOXSTEP(a,b,x)	(CLAMP (((x) - (a)) / ((b) - (a)), 0, 1))

typedef struct {
	float v[3];
} color_t;

float mod (float a, float b);
#ifndef min
float min (float a, float b);
#endif
#ifndef max
float max (float a, float b);
#endif
float abs (float a);
float clamp (float x, float a, float b);
float step (float a, float x);
float smoothstep (float a, float b, float x);
float linear_interpolate (float a, float b, float x);
float cosine_interpolate (float a, float b, float x);
float cubic_interpolate (float v0, float v1, float v2, float v3, float x);
float spline (float x, int nknots, float *knot);
float gammacorrect (float gamma, float x);
float bias (float b, float x);
float gain (float g, float x);
float boxstep (float a, float b, float x);

void mix (color_t C0, color_t C1, float f, color_t &out);

// Value Noise
void InitPermArray ();
float vnoise (float x, float y, float z);
void valueTabInit (int seed);

namespace Perlin {
	float Noise1 (float x);
	float Noise2 (float v[2]);
	float Noise2 (float x, float y);
	float Noise3 (float v[3]);
	float Noise3 (float x, float y, float z);
	void Init (void);
};

class PixelShader_t {
public:
	PixelShader_t ();
	virtual ~PixelShader_t ();

	virtual color_t GetColorAt (float x, float y, float z);
};

#endif