#ifndef FLUXIONS_SDLTEMPLATE_HPP
#define FLUXIONS_SDLTEMPLATE_HPP

#include <stdio.h>
#include <stdlib.h>

#include <GL/glew.h>
#include <SDL.h>
#include <SDL_image.h>

#include <viperfish/application.hpp>

#pragma comment(lib, "SDL.lib")
#pragma comment(lib, "SDLmain.lib")
#pragma comment(lib, "SDL_image.lib")
#pragma comment(lib, "glew32.lib")
#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "glu32.lib")

class SDLTemplate : public Viperfish::Application
{
public:
	SDLTemplate();
	~SDLTemplate();

	void SetVideoMode(int width, int height, int bpp, bool fullscreen);
	void SwapBuffers();
	void Quit();
	void SendQuitEvent();
	int GetWindowWidth() { return windowWidth; }
	int GetWindowHeight() { return windowHeight; }

	void Init();
	void Kill();
	void Start();
	void Stop();
	void Pause();
	void Resume();

	void RegisterApplication(Viperfish::Application *app);

	void OnInit();
	void OnKill();
	void OnUpdate(double deltaTime);
	void OnTimer(int timerId, int value, double timerInterval);
	void OnResize(int width, int height);
	void OnRender();
	void OnRenderOverlay();
	void OnStart();
	void OnStop();
	void OnPause();
	void OnResume();
	void OnKeyUp(int key);
	void OnKeyDown(int key);
	void OnSpecialKeyDown(int key);
	void OnSpecialKeyUp(int key);
	void OnMouseButtonDown(int button);
	void OnMouseButtonUp(int button);
	void OnMouseMotion(int x, int y, int xrel, int yrel, int buttons);
	void OnJoyAxis(int axis, float value, float delta);
	void OnJoyHat(int direction);
	void OnJoyBall(int ball, float value);
	void OnJoyButtonDown(int button);
	void OnJoyButtonUp(int button);
private:
	SDL_Window *mainWindow;
	SDL_GLContext mainContext;
	Viperfish::Application *application;
	int windowWidth, windowHeight;
	double t0, t1;

	// input state
	static const int MAX_JOYSTICKS = 8;
	int jaxis_value[MAX_JOYSTICKS];
	int jaxis_delta[MAX_JOYSTICKS];
	

	bool mainLoopRunning;
	bool mainLoopStarted;
	void MainEventLoop();
	void ExitFailure(const char *msg);
};


extern SDLTemplate SDL;

#endif
