// Fluxions 3D Engine
// simple_rendertotexture.hpp
// Copyright (C) 2000-2011 Jonathan Metzgar
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#ifndef FLUXIONS_SIMPLE_RENDERTOTEXTURE_HPP
#define FLUXIONS_SIMPLE_RENDERTOTEXTURE_HPP

#include <fluxions/opengl.hpp>

namespace Fluxions
{
	class SimpleRenderToTexture
	{
	public:
		// type is either GL_TEXTURE_2D or GL_TEXTURE_CUBE_MAP
		// buffers is GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT
		// width is the horizontal size of the frame buffer and textures
		// height is the vertical size of the frame buffer and textures
		SimpleRenderToTexture();
		void Init(GLenum type, GLbitfield buffers, int width, int height);
		void Reset();

		void Resize(int width, int height);
		void SetViewport(int x, int y, int w, int h);
		// for GL_TEXTURE_CUBE_MAP, texture = GL_TEXTURE_CUBE_MAP_[NEGATIVE|POSITIVE]_[X|Y|Z]
		// for GL_TEXTURE_2D, texture doesn't matter
		void BindCubeMapTexture(GLenum activeTexture, GLenum face, GLbitfield mask);
		void BindTexture(GLenum activeTexture, GLbitfield mask);
		void Clear(GLbitfield mask = 0);
		void BeginScene();
		void EndScene();
	private:
		GLenum textureType;
		GLint savedViewport[4];
		GLint viewport[4];
		GLint framebufferWidth;
		GLint framebufferHeight;
		GLuint framebuffer;
		GLuint depthBuffer, colorBuffer, depthStencilBuffer;
		GLuint colorTexture, depthTexture, depthStencilTexture;
		GLboolean usingColorBuffer;
		GLboolean usingDepthBuffer;
		GLboolean usingDepthStencilBuffer;
	};
}

#endif
