// Fluxions 3D Engine
// staticmodel.hpp
// Copyright (C) 2000-2011 Jonathan Metzgar
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#ifndef FLUXIONS_STATICMODEL_HPP
#define FLUXIONS_STATICMODEL_HPP

#include <string>
#include <vector>
#include <stdio.h>
#include <fluxions/ftl_math.hpp>
#include <fluxions/shaderprogram.hpp>
#include <fluxions/opengl.hpp>
#include <fluxions/bbox.hpp>

namespace Fluxions
{
	struct Vertex64
	{
		ftl::Vector3f position;
		ftl::Vector3f normal;
		ftl::Vector2f texcoord1;
		ftl::Vector4f texcoord2;
		ftl::Vector4f texcoord3;
	};

	struct Surface
	{
		GLenum mode;
		GLint first;
		GLsizei count;
		std::string materialName;
		std::string surfaceName;
		ShaderProgram *shaderProgram;

		Surface() : mode(GL_TRIANGLES), first(0), count(0), shaderProgram(NULL) {}
	};

	class StaticModel
	{
	public:
		StaticModel();
		~StaticModel();

		bool Save(const std::string &filename);
		bool Load(const std::string &filename);
		bool LoadXYZ(const std::string &filename, const float size = 1.0);
		bool SaveXYZ(const std::string &filename);
		bool LoadXYZ2(const std::string &filename);
		bool SaveXYZ2(const std::string &filename);
		bool SaveJS(const std::string &filename, const std::string &modelName);
		bool LoadOBJ(const std::string &filename, const double size = 1.0);
		bool SaveOBJ(const std::string &filename);

		void BindVertexArrays(ShaderProgram *shaderProgram);
		void UnbindVertexArrays(ShaderProgram *shaderProgram);
		void RenderSurface(int i);
		void SetShaderProgram(ShaderProgram *shaderProgram=NULL, int surface=-1);

		void Render(int surface=-1, bool fixedFunction=false);
		void RenderLines(int surface=-1);
		void RenderPoints(int surface=-1);
		void DebugRender(int surface=-1, GLenum drawMode = GL_FILL, bool drawNormals=false, bool useTexture=false, bool useLight=false);

		void ComputeTangentVectors();

		void Clear();
		void Resize(int vertexCount, int indexCount, int surfaceCount=1);
		void CreateSimpleModel(int vertexCount, int indexCount, int surfaceCount=1);
		void Compile();
		void Transform(const ftl::Matrix4d &mat);

		std::vector<Vertex64> Vertices;
		std::vector<GLuint> Indices;
		std::vector<Surface> Surfaces;
		Fluxions::BBox BoundingBox;

		static StaticModel CreateSuperquadric(double n, double e, double alpha, int segments, int slices, double ax, double ay, double az);
		static StaticModel CreateBox(float x1, float y1, float z1, float x2, float y2, float z2);
	private:
		GLuint arrayBufferID;
		GLuint elementArrayBufferID;

		int vertexBinding;
		int normalBinding;
		int texcoord1Binding;
		int texcoord2Binding;
		int texcoord3Binding;
	};
}

#endif
