#ifndef FLUXIONS_TEXTURE_HPP
#define FLUXIONS_TEXTURE_HPP


#include <fluxions/opengl.hpp>
#include <vector>
#include <string>


namespace Fluxions
{
	extern float Cubemap_PositiveX_Matrix[16];
	extern float Cubemap_NegativeX_Matrix[16];
	extern float Cubemap_PositiveY_Matrix[16];
	extern float Cubemap_NegativeY_Matrix[16];
	extern float Cubemap_PositiveZ_Matrix[16];
	extern float Cubemap_NegativeZ_Matrix[16];

	//mat4 cubemapPositiveXMatrix = mat4(0, 0, -1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1);
	//mat4 cubemapNegativeXMatrix = mat4(0, 0, 1, 0, 0, 1, 0, 0, -1, 0, 0, 0, 0, 0, 0, 1);
	//mat4 cubemapPositiveYMatrix = mat4(1, 0, 0, 0, 0, 0, -1, 0, 0, 1, 0, 0, 0, 0, 0, 1);
	//mat4 cubemapNegativeYMatrix = mat4(1, 0, 0, 0, 0, 0, 1, 0, 0, -1, 0, 0, 0, 0, 0, 1);
	//mat4 cubemapPositiveZMatrix = mat4(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1);
	//mat4 cubemapNegativeZMatrix = mat4(-1, 0, 0, 0, 0, 1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1);

	class SamplerObject
	{
	public:
		SamplerObject();
		~SamplerObject();

		void Create();
		void Delete();
		void Bind(GLuint unit);

		void SetWrapS(GLint param);
		void SetWrapT(GLint param);
		void SetWrapR(GLint param);
		void SetWrapST(GLint S, GLint T);
		void SetWrapTR(GLint T, GLint R);
		void SetWrapSTR(GLint S, GLint T, GLint R);
		void SetMinFilter(GLint minFilter);
		void SetMagFilter(GLint magFilter);
		void SetMinMagFilters(GLint minFilter, GLint magFilter);
		void SetCompareFunction(GLint func);
		void SetCompareMode(GLint mode);
		void SetMinLOD(GLfloat lod);
		void SetMaxLOD(GLfloat lod);

		void ApplySettings(GLuint textureUnit, GLuint texture, GLenum target);
		void Use(GLenum textureUnit, GLuint texture, GLenum target);

		GLuint GetId() const { return id; }
		GLint GetWrapS() const { return wrapS; }
		GLint GetWrapT() const { return wrapT; }
		GLint GetWrapR() const { return wrapR; }
		GLint GetMinFilter() const { return minFilter; }
		GLint GetMagFilter() const { return magFilter; }
		GLint GetCompareFunc() const { return compareFunc; }
		GLint GetCompareMode() const { return compareMode; }
		GLfloat GetMinLOD() const { return maxLOD; }
		GLfloat GetMaxLOD() const { return maxLOD; }
	private:
		GLint wrapS;
		GLint wrapT;
		GLint wrapR;
		GLint minFilter;
		GLint magFilter;
		GLint compareMode;
		GLint compareFunc;
		GLfloat minLOD;
		GLfloat maxLOD;
		GLuint id;
	};


	class Texture
	{
	public:
		Texture();
		~Texture();

		void Create();
		void Delete();
		void Bind(GLint unit, bool applySamplerObjectSettings = false);
		void Disable(GLint unit);
		void GenerateMipmaps();

		bool LoadTexture1D(const std::string &filename, bool genMipMap=true);
		bool LoadTexture2D(const std::string &filename, bool genMipMap=true);
		bool LoadTexture1DArray(const std::string &filename, bool genMipMap=true);
		bool LoadTexture2DArray(const std::vector<std::string> &filenames, bool genMipMap=true);
		bool LoadTexture2DArray(const std::string &filename, bool genMipMap=true);
		bool LoadTexture3D(const std::string &filename, int count, bool genMipMap=true);
		bool LoadTextureCubeMap(const std::vector<std::string> &filenames, bool genMipMap=true);

		void SetTexture1D(GLenum format, GLenum type, int width, void *data, bool genMipMap=true);
		void SetTexture2D(GLenum format, GLenum type, int width, int height, void *data, bool genMipMap=true);
		void SetTexture3D(GLenum format, GLenum type, int width, int height, int depth, void *data, bool genMipMap=true);
		void SetTexture1DArray(GLenum format, GLenum type, int width, int count, void *data, bool genMipMap=true);
		void SetTexture2DArray(GLenum format, GLenum type, int width, int height, int count, void *data, bool genMipMap=true);
		void SetTextureCubeMap(GLenum format, GLenum type, int width, int height, void **data, bool genMipMap=true);
		void SetTextureCubeMap(GLenum format, GLenum type, int width, int height, void *posxData, void *posyData, void *poszData, void *negxData, void *negyData, void *negzData, bool genMipMap=true);

		GLuint GetTextureId() const { return id; }
		GLenum GetTarget() const { return target; }

		SamplerObject samplerObject;
	private:
		GLuint id;
		GLenum target;
	};
}

#endif // FLUXIONS_TEXTURE_HPP
