#include <iostream>
#include <stdlib.h>
#include <fluxions/debug.hpp>
#include <SDL_image.h>
#include <GL/glew.h>
#include <GL/GL.h>
#include <GL/GLU.h>
#include <fluxions/texture.hpp>
#include <IL/il.h>

using namespace std;

namespace Fluxions
{
	float Cubemap_PositiveX_Matrix[16] = { 0, 0, -1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1 };
	float Cubemap_NegativeX_Matrix[16] = { 0, 0, 1, 0, 0, 1, 0, 0, -1, 0, 0, 0, 0, 0, 0, 1 };
	float Cubemap_PositiveY_Matrix[16] = { 1, 0, 0, 0, 0, 0, -1, 0, 0, 1, 0, 0, 0, 0, 0, 1 };
	float Cubemap_NegativeY_Matrix[16] = { 1, 0, 0, 0, 0, 0, 1, 0, 0, -1, 0, 0, 0, 0, 0, 1 };
	float Cubemap_PositiveZ_Matrix[16] = { 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1 };
	float Cubemap_NegativeZ_Matrix[16] = { -1, 0, 0, 0, 0, 1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1 };


	SamplerObject::SamplerObject()
	{
		wrapS = GL_REPEAT;
		wrapT = GL_REPEAT;
		wrapR = GL_REPEAT;
		minFilter = GL_NEAREST_MIPMAP_LINEAR;
		magFilter = GL_LINEAR;
		compareFunc = GL_LEQUAL;
		compareMode = GL_NONE;
		minLOD = -1000.0f;
		maxLOD = 1000.0f;
		id = 0;
	}


	SamplerObject::~SamplerObject()
	{
	}


	void SamplerObject::Create()
	{
		if (id != 0)
		{
			Delete();
		}
		glGenSamplers(1, &id);
	}


	void SamplerObject::Delete()
	{
		if (id != 0)
		{
			glDeleteSamplers(1, &id);
			id = 0;
		}
	}


	void SamplerObject::Bind(GLuint unit)
	{
		glBindSampler(unit, id);
	}


	void SamplerObject::SetWrapS(GLint param)
	{
		wrapS = param;
		if (id != 0) glSamplerParameteri(id, GL_TEXTURE_WRAP_S, param);
	}


	void SamplerObject::SetWrapT(GLint param)
	{
		wrapT = param;
		if (id != 0) glSamplerParameteri(id, GL_TEXTURE_WRAP_T, param);
	}


	void SamplerObject::SetWrapR(GLint param)
	{
		wrapR = param;
		if (id != 0) glSamplerParameteri(id, GL_TEXTURE_WRAP_R, param);
	}


	void SamplerObject::SetWrapST(GLint S, GLint T)
	{
		wrapS = S;
		wrapT = T;
		if (id != 0) glSamplerParameteri(id, GL_TEXTURE_WRAP_S, S);
		if (id != 0) glSamplerParameteri(id, GL_TEXTURE_WRAP_T, T);
	}


	void SamplerObject::SetWrapTR(GLint T, GLint R)
	{
		wrapT = T;
		wrapR = R;
		if (id != 0) glSamplerParameteri(id, GL_TEXTURE_WRAP_T, T);
		if (id != 0) glSamplerParameteri(id, GL_TEXTURE_WRAP_R, R);
	}


	void SamplerObject::SetWrapSTR(GLint S, GLint T, GLint R)
	{
		wrapS = S;
		wrapT = T;
		wrapR = R;
		if (id != 0) glSamplerParameteri(id, GL_TEXTURE_WRAP_S, S);
		if (id != 0) glSamplerParameteri(id, GL_TEXTURE_WRAP_T, T);
		if (id != 0) glSamplerParameteri(id, GL_TEXTURE_WRAP_R, R);
	}


	void SamplerObject::SetMinFilter(GLint param)
	{
		minFilter = param;
		if (id != 0) glSamplerParameteri(id, GL_TEXTURE_MIN_FILTER, minFilter);
	}


	void SamplerObject::SetMagFilter(GLint param)
	{
		magFilter = param;
		if (id != 0) glSamplerParameteri(id, GL_TEXTURE_MAG_FILTER, param);
	}


	void SamplerObject::SetMinMagFilters(GLint minFilter, GLint magFilter)
	{
		SetMinFilter(minFilter);
		SetMagFilter(magFilter);
	}


	void SamplerObject::SetCompareFunction(GLint func)
	{
		compareFunc = func;
		if (id != 0) glSamplerParameteri(id, GL_TEXTURE_COMPARE_FUNC, func);
	}


	void SamplerObject::SetCompareMode(GLint mode)
	{
		compareMode = mode;
		if (id != 0) glSamplerParameteri(id, GL_TEXTURE_COMPARE_MODE, mode);
	}


	void SamplerObject::SetMinLOD(GLfloat lod)
	{
		minLOD = lod;
		if (id != 0) glSamplerParameterf(id, GL_TEXTURE_MIN_LOD, lod);
	}


	void SamplerObject::SetMaxLOD(GLfloat lod)
	{
		maxLOD = lod;
		if (id != 0) glSamplerParameterf(id, GL_TEXTURE_MAX_LOD, lod);
	}


	void SamplerObject::ApplySettings(GLuint unit, GLuint texture, GLenum target)
	{
		if (unit >= GL_TEXTURE0 && unit <= GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS-1)
			glActiveTexture(unit);
		else
			glActiveTexture(GL_TEXTURE0 + unit);
		glBindTexture(target, texture);
		glTexParameteri(target, GL_TEXTURE_WRAP_S, wrapS);
		glTexParameteri(target, GL_TEXTURE_WRAP_T, wrapT);
		glTexParameteri(target, GL_TEXTURE_WRAP_R, wrapR);
		glTexParameteri(target, GL_TEXTURE_MIN_FILTER, minFilter);
		glTexParameteri(target, GL_TEXTURE_MAG_FILTER, magFilter);
		glTexParameterf(target, GL_TEXTURE_MIN_LOD, minLOD);
		glTexParameterf(target, GL_TEXTURE_MAX_LOD, maxLOD);
		glTexParameteri(target, GL_TEXTURE_COMPARE_FUNC, compareFunc);
		glTexParameteri(target, GL_TEXTURE_COMPARE_MODE, compareMode);
		glBindTexture(target, 0);
		glActiveTexture(GL_TEXTURE0);
	}


	void SamplerObject::Use(GLenum unit, GLuint texture, GLenum target)
	{
		if (unit >= GL_TEXTURE0 && unit <= GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS-1)
			glActiveTexture(unit);
		else
			glActiveTexture(GL_TEXTURE0 + unit);
		glBindTexture(target, texture);
		glTexParameteri(target, GL_TEXTURE_WRAP_S, wrapS);
		glTexParameteri(target, GL_TEXTURE_WRAP_T, wrapT);
		glTexParameteri(target, GL_TEXTURE_WRAP_R, wrapR);
		glTexParameteri(target, GL_TEXTURE_MIN_FILTER, minFilter);
		glTexParameteri(target, GL_TEXTURE_MAG_FILTER, magFilter);
		glTexParameterf(target, GL_TEXTURE_MIN_LOD, minLOD);
		glTexParameterf(target, GL_TEXTURE_MAX_LOD, maxLOD);
		glTexParameteri(target, GL_TEXTURE_COMPARE_FUNC, compareFunc);
		glTexParameteri(target, GL_TEXTURE_COMPARE_MODE, compareMode);
		glActiveTexture(GL_TEXTURE0);
	}


	// class Fluxions::Texture ///////////////////////////////////////


	Texture::Texture()
	{
		id = 0;
		target = GL_TEXTURE_2D;
	}


	Texture::~Texture()
	{
		Delete();
	}


	void Texture::Create()
	{
		if (id != 0)
			Delete();
		glGenTextures(1, &id);
	}


	void Texture::Delete()
	{
		if (id != 0)
		{
			glDeleteTextures(1, &id);
		}
		id = 0;
	}


	void Texture::Bind(GLint unit, bool applySamplerObjectSettings)
	{
		if (unit < GL_TEXTURE0)
			unit += GL_TEXTURE0;
		glActiveTexture(unit);
		glBindTexture(target, id);

		if (applySamplerObjectSettings != false)
		{
			samplerObject.ApplySettings(unit, id, target);
		}
		//glEnable(target);

		glActiveTexture(GL_TEXTURE0);
	}


	void Texture::Disable(GLint unit)
	{
		if (unit < GL_TEXTURE0)
			unit += GL_TEXTURE0;
		glActiveTexture(unit);
		//glDisable(target);
		glActiveTexture(GL_TEXTURE0);
	}

	
	void Texture::GenerateMipmaps()
	{
		glGenerateMipmap(target);
	}


	bool Texture::LoadTexture1D(const std::string &filename, bool genMipMap)
	{
		target = GL_TEXTURE_1D;

		SDL_Surface *image = IMG_Load(filename.c_str());
		if (image == nullptr)
			return false;

		int width = image->w;
		void *data = image->pixels;
		int format = image->format->BitsPerPixel == 24 ? GL_RGB :
			image->format->BitsPerPixel == 32 ? GL_RGBA : 0;
		if (format == 0)
			return false;

		Bind(0);
		SetTexture1D(format, GL_UNSIGNED_BYTE, width, data, genMipMap);
		glBindTexture(target, 0);
		SDL_FreeSurface(image);
		return true;
	}

	wstring StringToWString(const std::string &str)
	{
		wstring output;
		size_t strSize = strlen(str.c_str()) + 1;
		size_t convertedChars = 0;
		wchar_t *wcstring = new wchar_t[strSize];
		mbstowcs_s(&convertedChars, wcstring, strSize, str.c_str(), _TRUNCATE);
		output = wcstring;
		delete [] wcstring;
		return output;
	}

	bool Texture::LoadTexture2D(const std::string &filename, bool genMipMap)
	{
		target = GL_TEXTURE_2D;		

		SDL_Surface *image = IMG_Load(filename.c_str());
		if (image == NULL)
		{
			cout << "IMG_GetError() reports: " << IMG_GetError() << endl;
			return false;
		}

		int width = image->w;
		int height = image->h;
		void *data = image->pixels;
		int format = image->format->BitsPerPixel == 24 ? GL_RGB :
			image->format->BitsPerPixel == 32 ? GL_RGBA : 0;
		int internalformat = image->format->BitsPerPixel == 24 ? GL_RGB8 :
			image->format->BitsPerPixel == 32 ? GL_RGBA8 : 0;
		GLenum type = GL_UNSIGNED_BYTE;
		if (format == 0)
			return false;

		glGenTextures(1, &id);
		glBindTexture(GL_TEXTURE_2D, id);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, format, type, data);
		if (genMipMap != false)
		{
			glGenerateMipmap(target);
		}
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glBindTexture(GL_TEXTURE_2D, 0);
		SDL_FreeSurface(image);
		return true;
	}


	bool Texture::LoadTexture1DArray(const std::string &filename, bool genMipMap)
	{
		target = GL_TEXTURE_1D_ARRAY;

		SDL_Surface *image = IMG_Load(filename.c_str());
		if (image == nullptr)
			return false;

		int width = image->w;
		int height = image->h;
		void *data = image->pixels;
		int format = image->format->BitsPerPixel == 24 ? GL_RGB :
			image->format->BitsPerPixel == 32 ? GL_RGBA : 0;
		if (format == 0)
			return false;

		Bind(0);
		SetTexture2D(format, GL_UNSIGNED_BYTE, width, height, data, genMipMap);
		glBindTexture(target, 0);
		return true;
	}


	bool Texture::LoadTexture2DArray(const std::string &filename, bool genMipMap)
	{
		target = GL_TEXTURE_2D_ARRAY;
		Bind(0);

		if (genMipMap != false)
		{
			GenerateMipmaps();
		}
		glBindTexture(target, 0);
		return true;
	}


	bool Texture::LoadTexture2DArray(const std::vector<std::string> &filenames, bool genMipMap)
	{
		target = GL_TEXTURE_2D_ARRAY;

		unsigned char *data = nullptr;
		int count = filenames.size();
		int width = 0;
		int height = 0;
		int bpp = 0;
		int format = 0;
		bool badData = false;

		for (int i = 0; i < count; i++)
		{
			SDL_Surface *image = IMG_Load(filenames[i].c_str());
			if (image == nullptr)
			{
				badData = true;
				continue;
			}

			int imageFormat = image->format->BitsPerPixel == 24 ? GL_RGB :
				image->format->BitsPerPixel == 32 ? GL_RGBA : 0;
			if (imageFormat == 0)
				badData = true;
			if (i == 0)
			{
				width = image->w;
				height = image->h;
				bpp = image->format->BitsPerPixel / 8;
				format = imageFormat;

				data = new unsigned char[width * height * count * bpp];
			}
			else
			{
				if (image->w != width || image->h != height || format != imageFormat)
				{
					badData = true;
					break;
				}
			}
			if (imageFormat != 0)
			{
				memcpy(data + i * (width * height * bpp), image->pixels, width * height * bpp);
			}
			SDL_FreeSurface(image);
		}

		if (data == nullptr)
			return false;
		if (badData == true)
		{
			delete [] data;
			return false;
		}

		Bind(0);
		SetTexture3D(format, GL_UNSIGNED_BYTE, width, height, count, data, genMipMap);
		glBindTexture(target, 0);
		return true;
	}


	bool Texture::LoadTexture3D(const std::string &filename, int count, bool genMipMap)
	{
		target = GL_TEXTURE_3D;

		SDL_Surface *image = IMG_Load(filename.c_str());
		if (image == nullptr)
			return false;

		int width = image->w;
		int height = image->h / count;
		void *data = image->pixels;
		int format = image->format->BitsPerPixel == 24 ? GL_RGB :
			image->format->BitsPerPixel == 32 ? GL_RGBA : 0;
		if (format == 0)
			return false;

		Bind(0);
		SetTexture3D(format, GL_UNSIGNED_BYTE, width, height, count, data, genMipMap);
		glBindTexture(target, 0);
		return true;
	}


	bool Texture::LoadTextureCubeMap(const std::vector<std::string> &filenames, bool genMipMap)
	{
		target = GL_TEXTURE_CUBE_MAP;
		
		SDL_Surface *images[6];
		int width;
		int height;
		int format;
		int imageFormat;
		void *data[6];
		bool badData = false;
		for (int i = 0; i < 6; i++)
		{
			images[i] = IMG_Load(filenames[i].c_str());
			imageFormat = images[i]->format->BitsPerPixel == 24 ? GL_RGB :
				images[i]->format->BitsPerPixel == 32 ? GL_RGBA : 0;
			if (imageFormat == 0)
				badData = true;
			data[i] = images[i]->pixels;
			if (i == 0)
			{
				width = images[i]->w;
				height = images[i]->h;
				format = imageFormat;
			}
			else
			{
				if (width != images[i]->w || height != images[i]->h || format != imageFormat)
				{
					badData = true;
				}
			}
		}

		Create();
		Bind(0);
		SetTextureCubeMap(format, GL_UNSIGNED_BYTE, width, height, data, genMipMap);
		glBindTexture(target, 0);
		for (int i = 0; i < 6; i++)
		{
			SDL_FreeSurface(images[i]);
		}
		if (badData == true)
			return false;
		return true;
	}


	void Texture::SetTexture1D(GLenum format, GLenum type, int width, void *data, bool genMipMap)
	{
		glTexImage1D(target, 0, format, width, 0, format, type, data);
		if (genMipMap != false)
		{
			GenerateMipmaps();
		}
	}


	void Texture::SetTexture2D(GLenum format, GLenum type, int width, int height, void *data, bool genMipMap)
	{
		glTexImage2D(target, 0, format, width, height, 0, format, type, data);
		if (genMipMap != false)
		{
			GenerateMipmaps();
		}
	}


	void Texture::SetTexture3D(GLenum format, GLenum type, int width, int height, int depth, void *data, bool genMipMap)
	{
		glTexImage3D(target, 0, format, width, height, depth, 0, format, type, data);
		if (genMipMap != false)
		{
			GenerateMipmaps();
		}
	}


	void Texture::SetTexture1DArray(GLenum format, GLenum type, int width, int count, void *data, bool genMipMap)
	{
		glTexImage2D(target, 0, 0, width, count, 0, format, type, data);
		if (genMipMap != false)
		{
			GenerateMipmaps();
		}
	}


	void Texture::SetTexture2DArray(GLenum format, GLenum type, int width, int height, int count, void *data, bool genMipMap)
	{
		glTexImage3D(target, 0, format, width, height, count, 0, format, type, data);
		if (genMipMap != false)
		{
			GenerateMipmaps();
		}
	}


	void Texture::SetTextureCubeMap(GLenum format, GLenum type, int width, int height, void **data, bool genMipMap)
	{
		glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 0, format, width, height, 0, format, type, data[0]);
		glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X, 0, format, width, height, 0, format, type, data[1]);
		glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Y, 0, format, width, height, 0, format, type, data[2]);
		glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, 0, format, width, height, 0, format, type, data[3]);
		glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Z, 0, format, width, height, 0, format, type, data[4]);
		glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, 0, format, width, height, 0, format, type, data[5]);
		if (genMipMap != false)
		{
			GenerateMipmaps();
		}
	}


	void Texture::SetTextureCubeMap(GLenum format, GLenum type, int width, int height, void *posxData, void *posyData, void *poszData, void *negxData, void *negyData, void *negzData, bool genMipMap)
	{
		glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 0, format, width, height, 0, format, type, posxData);
		glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Y, 0, format, width, height, 0, format, type, posyData);
		glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Z, 0, format, width, height, 0, format, type, poszData);
		glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X, 0, format, width, height, 0, format, type, negxData);
		glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, 0, format, width, height, 0, format, type, negyData);
		glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, 0, format, width, height, 0, format, type, negzData);
		if (genMipMap != false)
		{
			GenerateMipmaps();
		}
	}


}
