#include <fluxions/shaderprogram.hpp>

namespace Fluxions
{
	using namespace ftl;
	using namespace std;

	Uniform::Uniform()
	{
	}


	Uniform::Uniform(const Uniform &u)
	{
		float_vector = u.float_vector;
		int_vector = u.int_vector;
		count = u.count;
		type = u.type;
	}


	Uniform::~Uniform()
	{
	}


	void Uniform::SetUniform(GLint loc)
	{
		switch (type)
		{
		case GL_FLOAT:
			glUniform1fv(loc, count, &float_vector[0]);
			break;
		case GL_FLOAT_VEC2:
			glUniform2fv(loc, count, &float_vector[0]);
			break;
		case GL_FLOAT_VEC3:
			glUniform3fv(loc, count, &float_vector[0]);
			break;
		case GL_FLOAT_VEC4:
			glUniform4fv(loc, count, &float_vector[0]);
			break;
		case GL_SAMPLER_1D:
		case GL_SAMPLER_2D:
		case GL_SAMPLER_3D:
		case GL_SAMPLER_CUBE:
		case GL_SAMPLER_1D_SHADOW:
		case GL_SAMPLER_2D_SHADOW:
		case GL_BOOL:
		case GL_INT:
			glUniform1iv(loc, count, &int_vector[0]);
			break;
		case GL_BOOL_VEC2:
		case GL_INT_VEC2:
			glUniform2iv(loc, count, &int_vector[0]);
			break;
		case GL_BOOL_VEC3:
		case GL_INT_VEC3:
			glUniform3iv(loc, count, &int_vector[0]);
			break;
		case GL_BOOL_VEC4:
		case GL_INT_VEC4:
			glUniform4iv(loc, count, &int_vector[0]);
			break;
		case GL_FLOAT_MAT2:
			glUniformMatrix2fv(loc, count, GL_FALSE, &float_vector[0]);
			break;
		case GL_FLOAT_MAT3:
			glUniformMatrix3fv(loc, count, GL_FALSE, &float_vector[0]);
			break;
		case GL_FLOAT_MAT4:
			glUniformMatrix4fv(loc, count, GL_FALSE, &float_vector[0]);
			break;
		case GL_FLOAT_MAT2x3:
			glUniformMatrix2x3fv(loc, count, GL_FALSE, &float_vector[0]);
			break;
		case GL_FLOAT_MAT2x4:
			glUniformMatrix2x4fv(loc, count, GL_FALSE, &float_vector[0]);
			break;
		case GL_FLOAT_MAT3x2:
			glUniformMatrix3x2fv(loc, count, GL_FALSE, &float_vector[0]);
			break;
		case GL_FLOAT_MAT3x4:
			glUniformMatrix3x4fv(loc, count, GL_FALSE, &float_vector[0]);
			break;
		case GL_FLOAT_MAT4x2:
			glUniformMatrix4x2fv(loc, count, GL_FALSE, &float_vector[0]);
			break;
		case GL_FLOAT_MAT4x3:
			glUniformMatrix4x3fv(loc, count, GL_FALSE, &float_vector[0]);
			break;
		}
	}

	void Uniform::operator=(const int &v) { IntegerVector(1, 1, &v); }
	void Uniform::operator=(const ftl::tvector2<int> &v) { IntegerVector(1, 2, v.v); }
	void Uniform::operator=(const ftl::tvector3<int> &v) { IntegerVector(1, 3, v.v); }
	void Uniform::operator=(const ftl::tvector4<int> &v) { IntegerVector(1, 4, v.v); }
	void Uniform::operator=(const float &v) { FloatVector(1, 1, &v); }
	void Uniform::operator=(const ftl::tvector2<float> &v) { FloatVector(1, 2, v.v); }
	void Uniform::operator=(const ftl::tvector3<float> &v) { FloatVector(1, 3, v.v); }
	void Uniform::operator=(const ftl::tvector4<float> &v) { FloatVector(1, 4, v.v); }
	void Uniform::operator=(const ftl::tmatrix2<float> &m) { Matrix2(1, m.m); }
	void Uniform::operator=(const ftl::tmatrix3<float> &m) { Matrix3(1, m.m); }
	void Uniform::operator=(const ftl::tmatrix4<float> &m) { Matrix4(1, m.m); }
	void Uniform::operator=(const ftl::tmatrix2x3<float> &m) { Matrix2x3(1, m.m); }
	void Uniform::operator=(const ftl::tmatrix2x4<float> &m) { Matrix2x4(1, m.m); }
	void Uniform::operator=(const ftl::tmatrix3x2<float> &m) { Matrix3x2(1, m.m); }
	void Uniform::operator=(const ftl::tmatrix3x4<float> &m) { Matrix3x4(1, m.m); }
	void Uniform::operator=(const ftl::tmatrix4x2<float> &m) { Matrix4x2(1, m.m); }
	void Uniform::operator=(const ftl::tmatrix4x3<float> &m) { Matrix4x3(1, m.m); }

	void Uniform::IntegerVector(int num, int size, const int *v)
	{
		if (size == 1)
			type = GL_INT;
		else if (size == 2)
			type = GL_INT_VEC2;
		else if (size == 3)
			type = GL_INT_VEC3;
		else if (size == 4)
			type = GL_INT_VEC4;
		count = num;
		int_vector.resize(num*size);
		copy(v, v+num*size, int_vector.begin());
	}

	void Uniform::BoolVector(int num, int size, const int *v)
	{
		if (size == 1)
			type = GL_BOOL;
		else if (size == 2)
			type = GL_BOOL_VEC2;
		else if (size == 3)
			type = GL_BOOL_VEC3;
		else if (size == 4)
			type = GL_BOOL_VEC4;
		count = num;
		int_vector.resize(num * size);
		copy(v, v+num*size, int_vector.begin());
	}

	void Uniform::FloatVector(int num, int size, const float *v)
	{
		if (size == 1)
			type = GL_FLOAT;
		else if (size == 2)
			type = GL_FLOAT_VEC2;
		else if (size == 3)
			type = GL_FLOAT_VEC3;
		else if (size == 4)
			type = GL_FLOAT_VEC4;
		count = num;
		float_vector.resize(num*size);
		copy(v, v+num*size, float_vector.begin());
	}

	void Uniform::Matrix2(int num, const float *v)
	{
		int size = 4;
		type = GL_FLOAT_MAT2;
		count = num;
		float_vector.resize(num*size);
		copy(v, v+num*size, float_vector.begin());
	}

	void Uniform::Matrix3(int num, const float *v)
	{
		int size = 9;
		type = GL_FLOAT_MAT3;
		count = num;
		float_vector.resize(num*size);
		copy(v, v+num*size, float_vector.begin());
	}

	void Uniform::Matrix4(int num, const float *v)
	{
		int size = 16;
		type = GL_FLOAT_MAT4;
		count = num;
		float_vector.resize(num*size);
		copy(v, v+num*size, float_vector.begin());
	}

	void Uniform::Matrix2x3(int num, const float *v)
	{
		int size = 6;
		type = GL_FLOAT_MAT2x3;
		count = num;
		float_vector.resize(num*size);
		copy(v, v+num*size, float_vector.begin());
	}

	void Uniform::Matrix2x4(int num, const float *v)
	{
		int size = 8;
		type = GL_FLOAT_MAT2x4;
		count = num;
		float_vector.resize(num*size);
		copy(v, v+num*size, float_vector.begin());
	}

	void Uniform::Matrix3x2(int num, const float *v)
	{
		int size = 6;
		type = GL_FLOAT_MAT3x2;
		count = num;
		float_vector.resize(num*size);
		copy(v, v+num*size, float_vector.begin());
	}

	void Uniform::Matrix3x4(int num, const float *v)
	{
		int size = 12;
		type = GL_FLOAT_MAT3x4;
		count = num;
		float_vector.resize(num*size);
		copy(v, v+num*size, float_vector.begin());
	}

	void Uniform::Matrix4x2(int num, const float *v)
	{
		int size = 8;
		type = GL_FLOAT_MAT4x2;
		count = num;
		float_vector.resize(num*size);
		copy(v, v+num*size, float_vector.begin());
	}

	void Uniform::Matrix4x3(int num, const float *v)
	{
		int size = 12;
		type = GL_FLOAT_MAT4x3;
		count = num;
		float_vector.resize(num*size);
		copy(v, v+num*size, float_vector.begin());
	}

	void Uniform::Sampler1D(int num, int *v)
	{
		type = GL_SAMPLER_1D;
		count = num;
		int_vector.resize(num);
		copy(v, v+num, int_vector.begin());
	}

	void Uniform::Sampler2D(int num, int *v)
	{
		type = GL_SAMPLER_2D;
		count = num;
		int_vector.resize(num);
		copy(v, v+num, int_vector.begin());
	}

	void Uniform::Sampler3D(int num, int *v)
	{
		type = GL_SAMPLER_3D;
		count = num;
		int_vector.resize(num);
		copy(v, v+num, int_vector.begin());
	}

	void Uniform::SamplerCube(int num, int *v)
	{
		type = GL_SAMPLER_CUBE;
		count = num;
		int_vector.resize(num);
		copy(v, v+num, int_vector.begin());
	}

	void Uniform::Sampler1DShadow(int num, int *v)
	{
		type = GL_SAMPLER_1D_SHADOW;
		count = num;
		int_vector.resize(num);
		copy(v, v+num, int_vector.begin());
	}

	void Uniform::Sampler2DShadow(int num, int *v)
	{
		type = GL_SAMPLER_2D_SHADOW;
		count = num;
		int_vector.resize(num);
		copy(v, v+num, int_vector.begin());
	}
}
