// Fluxions 3D Engine
// fluxions/glut_extensions.hpp
// Copyright (C) 2000-2011 Jonathan Metzgar
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

#include <algorithm>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include "fluxions/glut_extensions.hpp"
#include <SDL_image.h>


namespace GlutTemplatePrivate
{
	extern int g_screenWidth;
}


using namespace std;


void glutCornellBox(double scale)
{
	GLdouble floor[] = {
		552.8, 0.0, 0.0,
		0.0, 0.0, 0.0,
		0.0, 0.0, 559.2,
		549.6, 0.0, 559.2
	};

	GLdouble light[] = {
		343.0, 548.8, 227.0,
		343.0, 548.8, 332.0,
		213.0, 548.8, 332.0,
		213.0, 548.8, 227.0
	};

	GLdouble floor_holes[] = {
		130.0, 0.0,  65.0,
		 82.0, 0.0, 225.0,
		240.0, 0.0, 272.0,
		290.0, 0.0, 114.0,

		423.0, 0.0, 247.0,
		265.0, 0.0, 296.0,
		314.0, 0.0, 456.0,
		472.0, 0.0, 406.0
	};

	GLdouble ceiling[] = {
		556.0, 548.8, 0.0,   
		556.0, 548.8, 559.2,
		  0.0, 548.8, 559.2,
		  0.0, 548.8,   0.0
	};

	GLdouble ceiling_holes[] = {
		343.0, 548.8, 227.0,
		343.0, 548.8, 332.0,
		213.0, 548.8, 332.0,
		213.0, 548.8, 227.0
	};

	GLdouble back_wall[] = {
		549.6,   0.0, 559.2, 
		  0.0,   0.0, 559.2,
		  0.0, 548.8, 559.2,
		556.0, 548.8, 559.2	
	};

	GLdouble right_wall[] = {
		0.0,   0.0, 559.2,   
		0.0,   0.0,   0.0,
		0.0, 548.8,   0.0,
		0.0, 548.8, 559.2
	};

	GLdouble left_wall[] = {
		552.8,   0.0,   0.0, 
		549.6,   0.0, 559.2,
		556.0, 548.8, 559.2,
		556.0, 548.8,   0.0
	};

	GLdouble short_block[] = {
		130.0, 165.0,  65.0,
		 82.0, 165.0, 225.0,
		240.0, 165.0, 272.0,
		290.0, 165.0, 114.0,

		290.0,   0.0, 114.0,
		290.0, 165.0, 114.0,
		240.0, 165.0, 272.0,
		240.0,   0.0, 272.0,

		130.0,   0.0,  65.0,
		130.0, 165.0,  65.0,
		290.0, 165.0, 114.0,
		290.0,   0.0, 114.0,

		 82.0,   0.0, 225.0,
		 82.0, 165.0, 225.0,
		130.0, 165.0,  65.0,
		130.0,   0.0,  65.0,

		240.0,   0.0, 272.0,
		240.0, 165.0, 272.0,
		 82.0, 165.0, 225.0,
		 82.0,   0.0, 225.0
	};

	GLdouble tall_block[] = {
		423.0, 330.0, 247.0,
		265.0, 330.0, 296.0,
		314.0, 330.0, 456.0,
		472.0, 330.0, 406.0,

		423.0,   0.0, 247.0,
		423.0, 330.0, 247.0,
		472.0, 330.0, 406.0,
		472.0,   0.0, 406.0,

		472.0,   0.0, 406.0,
		472.0, 330.0, 406.0,
		314.0, 330.0, 456.0,
		314.0,   0.0, 456.0,

		314.0,   0.0, 456.0,
		314.0, 330.0, 456.0,
		265.0, 330.0, 296.0,
		265.0,   0.0, 296.0,

		265.0,   0.0, 296.0,
		265.0, 330.0, 296.0,
		423.0, 330.0, 247.0,
		423.0,   0.0, 247.0
	};
#define DRAWQUADS(ptr,offset,num)\
	for(int i = 0; i < num; i++)\
	{\
		glBegin(mode);\
		glVertex3dv(ptr+offset+i*12);\
		glVertex3dv(ptr+offset+i*12 + 3);\
		glVertex3dv(ptr+offset+i*12 + 6);\
		glVertex3dv(ptr+offset+i*12 + 9);\
		glEnd();\
		}
	//glTranslatef(-279.6, 0, 0);
	glRotatef(90.0, 1, 0, 0);
	glRotatef(180.0, 0, 1, 0);
	GLenum mode = GL_QUADS;
	glScaled(scale, scale, scale);
	glNormal3f(0, 0, 1);
	glColor3f(1, 1, 1);
	glBegin(mode);
	glVertex3dv(floor);
	glVertex3dv(floor+3);
	glVertex3dv(floor+6);
	glVertex3dv(floor+9);
	glEnd();
	glBegin(mode);
	glVertex3dv(floor_holes);
	glVertex3dv(floor_holes+3);
	glVertex3dv(floor_holes+6);
	glVertex3dv(floor_holes+9);
	glEnd();
	glBegin(mode);
	glVertex3dv(floor_holes+12);
	glVertex3dv(floor_holes+15);
	glVertex3dv(floor_holes+18);
	glVertex3dv(floor_holes+21);
	glEnd();
	glBegin(mode);
	glVertex3dv(light);
	glVertex3dv(light+3);
	glVertex3dv(light+6);
	glVertex3dv(light+9);
	glEnd();
	glBegin(mode);
	glVertex3dv(ceiling);
	glVertex3dv(ceiling+3);
	glVertex3dv(ceiling+6);
	glVertex3dv(ceiling+9);
	glEnd();
	glBegin(mode);
	glVertex3dv(ceiling_holes);
	glVertex3dv(ceiling_holes+3);
	glVertex3dv(ceiling_holes+6);
	glVertex3dv(ceiling_holes+9);
	glEnd();
	DRAWQUADS(back_wall, 0, 1);
	glNormal3f(0, 0, -1);
	glColor3f(0, 1, 0);
	DRAWQUADS(right_wall, 0, 1);
	glColor3f(1, 0, 0);
	DRAWQUADS(left_wall, 0, 1);
	glColor3f(1, 1, 1);
	DRAWQUADS(short_block, 0, 5);
	DRAWQUADS(tall_block, 0, 5);
}


void glutWireAxis(float size)
{
	glBegin(GL_LINES);
	glColor3f(1, 0, 0);
	glVertex3f(0, 0, 0); glVertex3f(size, 0, 0);
	glColor3f(0, 1, 0);
	glVertex3f(0, 0, 0); glVertex3f(0, size, 0);
	glColor3f(0, 0, 1);
	glVertex3f(0, 0, 0); glVertex3f(0, 0, size);
	glEnd();
}


void glutSprite(float x1, float y1, float x2, float y2, bool flipx, bool flipy)
{
	float minX = min(x1, x2);
	float maxX = max(x1, x2);
	float minY = min(y1, y2);
	float maxY = max(y1, y2);
	float s1 = flipx == false ? 0 : 1;
	float t1 = flipy == true ? 0 : 1;
	float s2 = flipx == false ? 1 : 0;
	float t2 = flipy == true ? 1 : 0;

	glBegin(GL_QUADS);
	glNormal3f(0, 0, 1);
	glTexCoord2f(s1, t1); glVertex2f(minX, minY);
	glTexCoord2f(s1, t2); glVertex2f(minX, maxY);
	glTexCoord2f(s2, t2); glVertex2f(maxX, maxY);
	glTexCoord2f(s2, t1); glVertex2f(maxX, minY);
	glEnd();
}


void glutSpriteRectangle(float x1, float y1, float x2, float y2, bool flipx, bool flipy)
{
	float minX = min(x1, x2);
	float maxX = max(x1, x2);
	float minY = min(y1, y2);
	float maxY = max(y1, y2);
	float s1 = flipx == false ? 0 : 1;
	float t1 = flipy == true ? 0 : 1;
	float s2 = flipx == false ? 1 : 0;
	float t2 = flipy == true ? 1 : 0;

	s1 *= (maxX - minX);
	s2 *= (maxX - minX);
	t1 *= (maxY - minY);
	t2 *= (maxY - minY);

	glBegin(GL_QUADS);
	glNormal3f(0, 0, 1);
	glTexCoord2f(s1, t1); glVertex2f(minX, minY);
	glTexCoord2f(s1, t2); glVertex2f(minX, maxY);
	glTexCoord2f(s2, t2); glVertex2f(maxX, maxY);
	glTexCoord2f(s2, t1); glVertex2f(maxX, minY);
	glEnd();
}


void glutSpriteRectangle(float x1, float y1, float x2, float y2, float width, float height, bool flipx, bool flipy)
{
	float minX = min(x1, x2);
	float maxX = max(x1, x2);
	float minY = min(y1, y2);
	float maxY = max(y1, y2);
	float s1 = flipx == false ? 0 : width;
	float t1 = flipy == true ? 0 : height;
	float s2 = flipx == false ? width : 0;
	float t2 = flipy == true ? height : 0;

	glBegin(GL_QUADS);
	glNormal3f(0, 0, 1);
	glTexCoord2f(s1, t1); glVertex2f(minX, minY);
	glTexCoord2f(s1, t2); glVertex2f(minX, maxY);
	glTexCoord2f(s2, t2); glVertex2f(maxX, maxY);
	glTexCoord2f(s2, t1); glVertex2f(maxX, minY);
	glEnd();
}


void glutSpriteVertexArray(float x1, float y1, float x2, float y2, int vloc, int tloc, bool flipx, bool flipy)
{
	float minX = min(x1, x2);
	float maxX = max(x1, x2);
	float minY = min(y1, y2);
	float maxY = max(y1, y2);
	float s1 = flipx == false ? 0 : 1;
	float t1 = flipy == true ? 0 : 1;
	float s2 = flipx == false ? 1 : 0;
	float t2 = flipy == true ? 1 : 0;

	GLfloat vertices[][3] = {
		{minX, minY, 0},
		{minX, maxY, 0},
		{maxX, maxY, 0},
		{maxX, minY, 0},
	};

	GLfloat texcoords[][2] = {
		{s1, t1},
		{s1, t2},
		{s2, t2},
		{s2, t1},
	};

	GLint indices[4] = {0, 1, 3, 2};

	if (vloc < 0)
	{
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glVertexPointer(3, GL_FLOAT, 0, vertices);
		glTexCoordPointer(2, GL_FLOAT, 0, texcoords);
		glDrawArrays(GL_QUADS, 0, 4);
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	}
	else
	{
		glVertexAttribPointer(vloc, 3, GL_FLOAT, GL_FALSE, 0, vertices);
		glEnableVertexAttribArray(vloc);
		if (tloc >= 0)
		{
			glVertexAttribPointer(tloc, 2, GL_FLOAT, GL_FALSE, 0, texcoords);
			glEnableVertexAttribArray(tloc);
		}
		glDrawElements(GL_TRIANGLE_STRIP, 4, GL_UNSIGNED_INT, indices);
		glDisableVertexAttribArray(vloc);
		if (tloc >= 0) glDisableVertexAttribArray(tloc);
	}
}


void glutSpriteProgram(GLuint program, float x1, float y1, float x2, float y2, bool flipx, bool flipy)
{
	glutSpriteProgram(program, "Vertex", "TexCoord", x1, y1, x2, y2, flipx, flipy);
}


void glutSpriteProgram(GLuint program, const char *vertexAttrib, const char *texCoordAttrib, float x1, float y1, float x2, float y2, bool flipx, bool flipy)
{
	float minX = min(x1, x2);
	float maxX = max(x1, x2);
	float minY = min(y1, y2);
	float maxY = max(y1, y2);
	float s1 = flipx == false ? 0 : 1;
	float t1 = flipy == true ? 0 : 1;
	float s2 = flipx == false ? 1 : 0;
	float t2 = flipy == true ? 1 : 0;

	GLfloat vertices[][3] = {
		{minX, minY, 0},
		{minX, maxY, 0},
		{maxX, maxY, 0},
		{maxX, minY, 0},
	};

	GLfloat texcoords[][2] = {
		{s1, t1},
		{s1, t2},
		{s2, t2},
		{s2, t1},
	};

	GLint indices[4] = {0, 1, 3, 2};

	int vloc = -1;
	int tloc = -1;
	if (glIsProgram(program) != GL_FALSE)
	{
		vloc = glGetAttribLocation(program, vertexAttrib);
		tloc = glGetAttribLocation(program, texCoordAttrib);
		glUseProgram(program);
	}

	if (vloc < 0)
	{
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glVertexPointer(3, GL_FLOAT, 0, vertices);
		glTexCoordPointer(2, GL_FLOAT, 0, texcoords);
		glDrawArrays(GL_QUADS, 0, 4);
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	}
	else
	{
		glVertexAttribPointer(vloc, 3, GL_FLOAT, GL_FALSE, 0, vertices);
		glEnableVertexAttribArray(vloc);
		if (tloc >= 0)
		{
			glVertexAttribPointer(tloc, 2, GL_FLOAT, GL_FALSE, 0, texcoords);
			glEnableVertexAttribArray(tloc);
		}
		glDrawElements(GL_TRIANGLE_STRIP, 4, GL_UNSIGNED_INT, indices);
		glDisableVertexAttribArray(vloc);
		if (tloc >= 0) glDisableVertexAttribArray(tloc);
	}
}


void glutWireBox(float x1, float y1, float z1, float x2, float y2, float z2)
{
	float minX = min(x1, x2);
	float maxX = max(x1, x2);
	float minY = min(y1, y2);
	float maxY = max(y1, y2);
	float minZ = min(z1, z2);
	float maxZ = max(z1, z2);
	
	glBegin(GL_LINE_LOOP);
	glNormal3f(0, -1, 0);
	glTexCoord2f(1, 0); glVertex3f(minX, minY, minZ);
	glTexCoord2f(0, 0); glVertex3f(maxX, minY, minZ);
	glTexCoord2f(0, 1); glVertex3f(maxX, minY, maxZ);
	glTexCoord2f(1, 1); glVertex3f(minX, minY, maxZ);
	glEnd();
	
	
	glBegin(GL_LINE_LOOP);
	glNormal3f(0, 1, 0);
	glTexCoord2f(0, 0); glVertex3f(minX, maxY, minZ);
	glTexCoord2f(1, 0); glVertex3f(maxX, maxY, minZ);
	glTexCoord2f(1, 1); glVertex3f(maxX, maxY, maxZ);
	glTexCoord2f(0, 1); glVertex3f(minX, maxY, maxZ);
	glEnd();
	
	
	glBegin(GL_LINE_LOOP);
	glNormal3f(-1, 0, 0);
	glTexCoord2f(0, 1); glVertex3f(minX, minY, minZ);
	glTexCoord2f(0, 0); glVertex3f(minX, maxY, minZ);
	glTexCoord2f(1, 0); glVertex3f(minX, maxY, maxZ);
	glTexCoord2f(1, 1); glVertex3f(minX, minY, maxZ);
	glEnd();
	
	
	glBegin(GL_LINE_LOOP);
	glNormal3f(1, 0, 0);
	glTexCoord2f(1, 1); glVertex3f(maxX, minY, minZ);
	glTexCoord2f(1, 0); glVertex3f(maxX, maxY, minZ);
	glTexCoord2f(0, 0); glVertex3f(maxX, maxY, maxZ);
	glTexCoord2f(0, 1); glVertex3f(maxX, minY, maxZ);
	glEnd();
	
	
	glBegin(GL_LINE_LOOP);
	glNormal3f(0, 0, -1);
	glTexCoord2f(1, 1); glVertex3f(minX, minY, minZ);
	glTexCoord2f(0, 1); glVertex3f(maxX, minY, minZ);
	glTexCoord2f(0, 0); glVertex3f(maxX, maxY, minZ);
	glTexCoord2f(1, 0); glVertex3f(minX, maxY, minZ);
	glEnd();
	
	
	glBegin(GL_LINE_LOOP);
	glNormal3f(0, 0, 1);
	glTexCoord2f(0, 1); glVertex3f(minX, minY, maxZ);
	glTexCoord2f(1, 1); glVertex3f(maxX, minY, maxZ);
	glTexCoord2f(1, 0); glVertex3f(maxX, maxY, maxZ);
	glTexCoord2f(0, 0); glVertex3f(minX, maxY, maxZ);	
	glEnd();
}


void glutSolidBox(float x1, float y1, float z1, float x2, float y2, float z2)
{
	float minX = min(x1, x2);
	float maxX = max(x1, x2);
	float minY = min(y1, y2);
	float maxY = max(y1, y2);
	float minZ = min(z1, z2);
	float maxZ = max(z1, z2);
	
	glBegin(GL_QUADS);
	
	glNormal3f(0, -1, 0);
	glTexCoord2f(1, 0); glVertex3f(minX, minY, minZ);
	glTexCoord2f(0, 0); glVertex3f(maxX, minY, minZ);
	glTexCoord2f(0, 1); glVertex3f(maxX, minY, maxZ);
	glTexCoord2f(1, 1); glVertex3f(minX, minY, maxZ);
	
	
	glNormal3f(0, 1, 0);
	glTexCoord2f(0, 0); glVertex3f(minX, maxY, minZ);
	glTexCoord2f(0, 1); glVertex3f(minX, maxY, maxZ);
	glTexCoord2f(1, 1); glVertex3f(maxX, maxY, maxZ);
	glTexCoord2f(1, 0); glVertex3f(maxX, maxY, minZ);
	
	
	glNormal3f(-1, 0, 0);
	glTexCoord2f(0, 1); glVertex3f(minX, minY, minZ);
	glTexCoord2f(1, 1); glVertex3f(minX, minY, maxZ);
	glTexCoord2f(1, 0); glVertex3f(minX, maxY, maxZ);
	glTexCoord2f(0, 0); glVertex3f(minX, maxY, minZ);
	
	
	glNormal3f(1, 0, 0);
	glTexCoord2f(1, 1); glVertex3f(maxX, minY, minZ);
	glTexCoord2f(1, 0); glVertex3f(maxX, maxY, minZ);
	glTexCoord2f(0, 0); glVertex3f(maxX, maxY, maxZ);
	glTexCoord2f(0, 1); glVertex3f(maxX, minY, maxZ);
	
	
	glNormal3f(0, 0, -1);
	glTexCoord2f(1, 1); glVertex3f(minX, minY, minZ);
	glTexCoord2f(1, 0); glVertex3f(minX, maxY, minZ);
	glTexCoord2f(0, 0); glVertex3f(maxX, maxY, minZ);
	glTexCoord2f(0, 1); glVertex3f(maxX, minY, minZ);
	

	glNormal3f(0, 0, 1);
	glTexCoord2f(0, 1); glVertex3f(minX, minY, maxZ);
	glTexCoord2f(1, 1); glVertex3f(maxX, minY, maxZ);
	glTexCoord2f(1, 0); glVertex3f(maxX, maxY, maxZ);
	glTexCoord2f(0, 0); glVertex3f(minX, maxY, maxZ);
	
	glEnd();
}


void glutSolidBoxCM(float x1, float y1, float z1, float x2, float y2, float z2, bool setcolors)
{
	float minX = min(x1, x2);
	float maxX = max(x1, x2);
	float minY = min(y1, y2);
	float maxY = max(y1, y2);
	float minZ = min(z1, z2);
	float maxZ = max(z1, z2);
	
	glBegin(GL_QUADS);
	
	if (setcolors) glColor3f(1, 0, 1);
	glNormal3f(0, -1, 0);
	glTexCoord3f(minX, minY, minZ);
	glVertex3f(minX, minY, minZ);
	glTexCoord3f(maxX, minY, minZ);
	glVertex3f(maxX, minY, minZ);
	glTexCoord3f(maxX, minY, maxZ);
	glVertex3f(maxX, minY, maxZ);
	glTexCoord3f(minX, minY, maxZ);
	glVertex3f(minX, minY, maxZ);
	
	
	if (setcolors) glColor3f(0, 1, 0);
	glNormal3f(0, 1, 0);
	glTexCoord3f(minX, maxY, minZ);
	glVertex3f(minX, maxY, minZ);
	glTexCoord3f(minX, maxY, maxZ);
	glVertex3f(minX, maxY, maxZ);
	glTexCoord3f(maxX, maxY, maxZ);
	glVertex3f(maxX, maxY, maxZ);
	glTexCoord3f(maxX, maxY, minZ);
	glVertex3f(maxX, maxY, minZ);
	
	
	if (setcolors) glColor3f(0, 1, 1);
	glNormal3f(-1, 0, 0);
	glTexCoord3f(minX, minY, minZ);
	glVertex3f(minX, minY, minZ);
	glTexCoord3f(minX, minY, maxZ);
	glVertex3f(minX, minY, maxZ);
	glTexCoord3f(minX, maxY, maxZ);
	glVertex3f(minX, maxY, maxZ);
	glTexCoord3f(minX, maxY, minZ);
	glVertex3f(minX, maxY, minZ);
	
	
	if (setcolors) glColor3f(1, 0, 0);
	glNormal3f(1, 0, 0);
	glTexCoord3f(maxX, minY, minZ);
	glVertex3f(maxX, minY, minZ);
	glTexCoord3f(maxX, maxY, minZ);
	glVertex3f(maxX, maxY, minZ);
	glTexCoord3f(maxX, maxY, maxZ);
	glVertex3f(maxX, maxY, maxZ);
	glTexCoord3f(maxX, minY, maxZ);
	glVertex3f(maxX, minY, maxZ);
	
	
	if (setcolors) glColor3f(1, 1, 0);
	glNormal3f(0, 0, -1);
	glTexCoord3f(minX, minY, minZ);
	glVertex3f(minX, minY, minZ);
	glTexCoord3f(minX, maxY, minZ);
	glVertex3f(minX, maxY, minZ);
	glTexCoord3f(maxX, maxY, minZ);
	glVertex3f(maxX, maxY, minZ);
	glTexCoord3f(maxX, minY, minZ);
	glVertex3f(maxX, minY, minZ);
	
	if (setcolors) glColor3f(0, 0, 1);
	glNormal3f(0, 0, 1);
	glTexCoord3f(minX, minY, maxZ);
	glVertex3f(minX, minY, maxZ);
	glTexCoord3f(maxX, minY, maxZ);
	glVertex3f(maxX, minY, maxZ);
	glTexCoord3f(maxX, maxY, maxZ);
	glVertex3f(maxX, maxY, maxZ);
	glTexCoord3f(minX, maxY, maxZ);
	glVertex3f(minX, maxY, maxZ);
	
	glEnd();
}


void glutCheckerboard(int wsegments, int hsegments, const float *whiteColor, const float *blackColor, const float xscale, float yscale, const float xtrans, const float ytrans)
{
	glBegin(GL_QUADS);
	for (int i = 0; i < wsegments; i++)
	{
		for (int j = 0; j < hsegments; j++)
		{
			glNormal3f(0, 0, 1);
			if ((j % 2 && i % 2) || j == i || (!(j % 2) && !(i % 2)))
				glColor4fv(whiteColor);
			else
				glColor4fv(blackColor);
				
			float x1, y1, x2, y2;
			x1 = (i) / (float)wsegments;
			y1 = (j) / (float)hsegments;
			x2 = ((i+1)) / (float)wsegments;
			y2 = ((j+1)) / (float)hsegments;
			float s1, t1, s2, t2;
			s1 = x1; t1 = y1;
			s2 = x2; t2 = y2;
			glTexCoord2f(s1, t1); glVertex2f(x1*xscale+xtrans, y1*yscale+ytrans);
			glTexCoord2f(s2, t1); glVertex2f(x2*xscale+xtrans, y1*yscale+ytrans);
			glTexCoord2f(s2, t2); glVertex2f(x2*xscale+xtrans, y2*yscale+ytrans);
			glTexCoord2f(s1, t2); glVertex2f(x1*xscale+xtrans, y2*yscale+ytrans);
		}
	}
	glEnd();
}


void glutCheckerboard(const float *whiteColor, const float *blackColor, const float xscale, float yscale, const float xtrans, const float ytrans)
{
	glBegin(GL_QUADS);
	for (int i = -4; i < 4; i++)
	{
		for (int j = -4; j < 4; j++)
		{
			glNormal3f(0, 0, 1);
			if ((j % 2 && i % 2) || j == i || (!(j % 2) && !(i % 2)))
				glColor4fv(whiteColor);
			else
				glColor4fv(blackColor);
				
			float x1, y1, x2, y2;
			x1 = (4 + i) / 8.0f;
			y1 = (4 + j) / 8.0f;
			x2 = (4 + (i+1)) / 8.0f;
			y2 = (4 + (j+1)) / 8.0f;
			//glRectf(x1, y1, x2, y2);
			float s1, t1, s2, t2;
			s1 = x1; t1 = y1; s2 = x2; t2 = y2;
			glTexCoord2f(s1, t1); glVertex2f(x1*xscale+xtrans, y1*yscale+ytrans);
			glTexCoord2f(s2, t1); glVertex2f(x2*xscale+xtrans, y1*yscale+ytrans);
			glTexCoord2f(s2, t2); glVertex2f(x2*xscale+xtrans, y2*yscale+ytrans);
			glTexCoord2f(s1, t2); glVertex2f(x1*xscale+xtrans, y2*yscale+ytrans);
		}
	}
	glEnd();
}


void glutLine2(float x1, float y1, float x2, float y2)
{
	glBegin(GL_LINES);
	glVertex2f(x1, y1);
	glVertex2f(x2, y2);
	glEnd();
}


void glutLine3(float x1, float y1, float z1, float x2, float y2, float z2)
{
	glBegin(GL_LINES);
	glVertex3f(x1, y1, z1);
	glVertex3f(x2, y2, z2);
	glEnd();
}


void glutBitmapString(void *font, const char *string)
{
	int len = strlen(string);
	int i = 0;
	
	for (i = 0; i < len; i++)
	{
		glutBitmapCharacter(font, string[i]);
	}	
}


void glutStrokeString(void *font, const char *string)
{
	int len = strlen(string);
	int i = 0;
	
	for (i = 0; i < len; i++)
	{
		glutStrokeCharacter(font, string[i]);
	}	
}


void glutBitmapPrintRight(void *font, int y, const char *format, ...)
{
	char buffer[256];
	va_list va;
	va_start(va, format);
	vsnprintf(buffer, 256, format, va);
	int width = glutBitmapLength(font, (unsigned char *)buffer);
	glRasterPos2f(GlutTemplatePrivate::g_screenWidth - width, y);
	glutBitmapString(font, buffer);
}


void glutBitmapPrintLeft(void *font, int y, const char *format, ...)
{
	char buffer[256];
	va_list va;
	va_start(va, format);
	vsnprintf(buffer, 256, format, va);
	glRasterPos2f(0, y);
	glutBitmapString(font, buffer);
}


void glutBitmapPrintCenter(void *font, int y, const char *format, ...)
{
	char buffer[256];
	va_list va;
	va_start(va, format);
	vsnprintf(buffer, 256, format, va);
	int width = glutBitmapLength(font, (unsigned char *)buffer);
	glRasterPos2f((GlutTemplatePrivate::g_screenWidth - width) / 2.0, y);
	glutBitmapString(font, buffer);
}


void glutBitmapPrint(void *font, int x, int y, const char *format, ...)
{
	char buffer[256];
	va_list va;
	va_start(va, format);
	vsnprintf(buffer, 256, format, va);
	glRasterPos2f(x, y);
	glutBitmapString(font, buffer);
}


void glutStrokePrintRight(void *font, int y, const char *format, ...)
{
	char buffer[256];
	va_list va;
	va_start(va, format);
	vsnprintf(buffer, 256, format, va);
	int width = glutStrokeLength(font, (unsigned char *)buffer);
	glRasterPos2f(GlutTemplatePrivate::g_screenWidth - width, y);
	glutStrokeString(font, buffer);
}


void glutStrokePrintLeft(void *font, int y, const char *format, ...)
{
	char buffer[256];
	va_list va;
	va_start(va, format);
	vsnprintf(buffer, 256, format, va);
	glRasterPos2f(0, y);
	glutStrokeString(font, buffer);
}


void glutStrokePrintCenter(void *font, int y, const char *format, ...)
{
	char buffer[256];
	va_list va;
	va_start(va, format);
	vsnprintf(buffer, 256, format, va);
	int width = glutStrokeLength(font, (unsigned char *)buffer);
	glRasterPos2f((GlutTemplatePrivate::g_screenWidth - width) / 2.0, y);
	glutStrokeString(font, buffer);
}


void glutStrokePrint(void *font, int x, int y, const char *format, ...)
{
	char buffer[256];
	va_list va;
	va_start(va, format);
	vsnprintf(buffer, 256, format, va);
	glRasterPos2f(x, y);
	glutStrokeString(font, buffer);
}


//////////////////////////////////////////////////////////////////////
// O P E N G L   E X T E N S I O N   F U N C T I O N S ///////////////
//////////////////////////////////////////////////////////////////////


void gluPointLight(GLenum light, GLfloat x, GLfloat y, GLfloat z)
{
	GLfloat position[4] = {x, y, z, 1};
	glLightfv(light, GL_POSITION, position);
}


void gluMaterial(GLfloat *ambient, GLfloat *diffuse, GLfloat *specular)
{
	glMaterialfv(GL_FRONT, GL_AMBIENT, ambient);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse);
	glMaterialfv(GL_FRONT, GL_SPECULAR, specular);
}

GLuint CreateTexture()
{
	GLuint texture;
	glGenTextures(1, &texture);
	return texture;
}

void DeleteTexture(GLuint texID)
{
	GLuint texture = texID;
	glDeleteTextures(1, &texture);
}

ftl::Matrix4f GetOpenGLMatrixf(GLenum matrix)
{
	GLfloat m[16];
	glGetFloatv(matrix, m);
	return ftl::Matrix4f(m);
}


ftl::Matrix4d GetOpenGLMatrixd(GLenum matrix)
{
	GLdouble m[16];
	glGetDoublev(matrix, m);
	return ftl::Matrix4d(m);
}


GLuint LoadTexture(const char *file, GLenum target)
{
	printf("LoadTexture(): loading texture from file: %s...", file);

	if (target == GL_TEXTURE_1D)
		return LoadTexture1D(file);
	GLuint texture = 0;
	SDL_Surface *surface = IMG_Load(file);
	if (!surface)
		return 0;
	if (surface->format->BitsPerPixel != 24 && surface->format->BitsPerPixel != 32)
	{
		SDL_FreeSurface(surface);
		return 0;
	}
	GLint internalformat = 0;
	GLenum format = 0;
	if (surface->format->BitsPerPixel == 24)
	{
		internalformat = GL_RGB8;
		format = GL_RGB;
	}
	else if (surface->format->BitsPerPixel == 32)
	{
		internalformat = GL_RGBA8;
		format = GL_RGBA;
	}
	glGenTextures(1, &texture);
	glBindTexture(target, texture);
	glTexImage2D(target, 0, internalformat, surface->w, surface->h, 0, format, GL_UNSIGNED_BYTE, surface->pixels);
	printf("%s\n", glGetError() == GL_NO_ERROR ? "success!" : "failure!");

	if (target != GL_TEXTURE_RECTANGLE)
	{
		glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glGenerateMipmap(target);
	}
	else
	{
		glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_CLAMP);
	}
	glBindTexture(target, 0);

	SDL_FreeSurface(surface);
	return texture;
}


GLuint LoadTexture1D(const char *file)
{
	GLuint texture = 0;
	FILE *fin = fopen(file, "rb");
	if (!fin)
		return 0;
	fseek(fin, 0, SEEK_END);
	int len = ftell(fin);
	fseek(fin, 0, SEEK_SET);
	GLbyte *data = new GLbyte[len];
	fread(data, 1, len, fin);
	fclose(fin);
	int width = len / 3;
	if (strcmp(file, "data/PAL1.PAL")==0)
		for(int i = 0; i < len; i++)
		{
			data[i] *= 4;
		}
	//if (strcmp(file, "data/PAL2.PAL")==0)
	//{
	//	FILE *fout=fopen("data/PAL2.PAL", "wb");
	//	for (int i = 0; i < 256; i++)
	//	{
	//		int intensity = i;
	//		int r = min(i, 85) * 3;
	//		intensity = max(intensity-85, 0);
	//		int g = min(intensity, 85) * 3;
	//		intensity = max(intensity-85, 0);
	//		int b = min(intensity, 85) * 3;
	//		data[i*3] = (char)r;
	//		data[i*3+1] = (char)g;
	//		data[i*3+2] = (char)b;
	//	}
	//	fwrite(data, 1, 768, fout);
	//	fclose(fout);
	//}
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_1D, texture);
	glTexImage1D(GL_TEXTURE_1D, 0, GL_RGB, width, 0, GL_RGB, GL_UNSIGNED_BYTE, data);
	//gluBuild1DMipmaps(GL_TEXTURE_1D, 4, width, GL_RGBA, GL_UNSIGNED_BYTE, data);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glGenerateMipmap(GL_TEXTURE_1D);
	glBindTexture(GL_TEXTURE_1D, 0);
	delete [] data;
	return texture;
}


void glutRenderQuickTeapot()
{
	glEnable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluPerspective(45.0f, 1.0, 0.001, 100.0);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glTranslatef(0, 0, -5);
	glColor3f(0.2, 0.3, 0.4);
	glutSolidTeapot(1.0);
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glColor3f(0.0, 0.0, 0.0);
	glDisable(GL_DEPTH_TEST);
}


void glutRenderQuickOrthoRect()
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(-1.0, 1.0, -1.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glColor3f(0.2, 0.3, 0.4);
	glRectf(-0.5, -0.5, 0.5, 0.5);
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glColor3f(0.0, 0.0, 0.0);
}
