#ifndef LIBMFG_KASL_HPP
#define LIBMFG_KASL_HPP

#include <map>
#include <list>
#include <vector>
#include <string>

struct KASL_Variable
{
	std::string name;
	std::string string;
	float value;
};

struct KASL_Command
{
	std::string name;
	std::string help;
	void (*callback)(std::vector<std::string> &, void*);
	void *data;
};

class KASL_VarCmdProcessor
{
public:
	KASL_VarCmdProcessor();
	~KASL_VarCmdProcessor();

	bool IsVariable(const char *str);
	KASL_Variable *GetVariable(const char *str);
	const char *PartialMatchVariableName(const char *str);
	void AddVariable(const char *str, float value);
	void AddVariable(const char *str, const char *string);
	void DeleteVariable(const char *str);
	void SetVariableValue(const char *str, float value);
	void SetVariableValue(const char *str, const char *value);
	float GetVariableValue(const char *str);
	std::string GetVariableString(const char *str);

	void TokenizeString(const char *str, std::vector<std::string> &tokens);

	bool IsCommand(const char *str);
	KASL_Command *GetCommand(const char *str);
	const char *PartialMatchCommandName(const char *str);
	void AddCommand(const char *str, void (*func)(std::vector<std::string> &, void*), void *data=NULL);
	void DeleteCommand(const char *str);
	bool Execute(const char *str);
private:
	std::list<KASL_Variable> vars;
	std::list<KASL_Command> cmds;
};

#endif
