#include "libmfg_tree_node.hpp"


namespace libmfg
{
	using namespace std;


	tree_node::tree_node()
	{
	}
	
	
	tree_node::~tree_node()
	{
		while (!empty())
		{
			remove_child(0);
		}
	}
	
	
	void tree_node::attach_child(tree_leaf_ptr child)
	{
		tree_node_ptr self = dynamic_pointer_cast<tree_node>(shared_from_this());		
		insert_child(child);
		child->set_parent(self);
	}
	
	
	void tree_node::insert_child(tree_leaf_ptr child)
	{
		tree_node_ptr self = dynamic_pointer_cast<tree_node>(shared_from_this());		
		children_.push_back(child);
		child->set_parent(self);
	}
	
	
	void tree_node::insert_child(int index, tree_leaf_ptr child)
	{
		list<tree_leaf_ptr>::iterator it = children_.begin();
		while (index--)
		{
			++it;
		}
		children_.insert(it, child);	
	}
	
	
	void tree_node::remove_child(tree_leaf_ptr child)
	{
		children_.remove(child);
		child->remove_parent();
	}
	
	
	void tree_node::remove_child(int index)
	{
		list<tree_leaf_ptr>::iterator it = children_.begin();
		
		if (index >= (int)children_.size())
		{
			return;
		}
		
		while (index--)
		{
			++it;
		}
		(*it)->remove_parent();
		(*it)->detach();
		children_.erase(it);
	}


	list<tree_leaf_ptr>::iterator tree_node::begin()
	{
		return children_.begin();
	}


	list<tree_leaf_ptr>::iterator tree_node::end()
	{
		return children_.end();
	}
	
	
	int tree_node::get_num_children() const
	{
		return (int)children_.size();
	}
	
	
	tree_leaf_ptr tree_node::get_child(int index)
	{
		list<tree_leaf_ptr>::iterator it = children_.begin();
		while (index--)
		{
			++it;
		}
		
		return *it;
	}
	
	
	bool tree_node::empty()
	{
		return children_.empty();
	}
	
	
	bool tree_node::is_child(tree_leaf_ptr child)
	{
		return false;
	}
	
	
	bool tree_node::is_child(int index)
	{
		if (index >= 0 && index < (int)children_.size())
		{
			return false;
		}
		return true;
	}
}
