#include <iostream>
#include <fluxions/sdltemplate.hpp>
#include <fluxions/opengl.hpp>
#include <IL/ilut.h>

using namespace std;


SDLTemplate::SDLTemplate()
{
	mainWindow = NULL;
	application = NULL;
	mainLoopRunning = false;
	mainLoopStarted = false;
	windowWidth = 0;
	windowHeight = 0;
}


SDLTemplate::~SDLTemplate()
{
	if (mainWindow != NULL)
		Quit();
}


void SDLTemplate::SetVideoMode(int width, int height, int bpp, bool fullscreen)
{
	if (SDL_Init(SDL_INIT_VIDEO) < 0)
		ExitFailure("Unable to initialize SDL");

	int flags = SDL_WINDOW_OPENGL | SDL_WINDOW_SHOWN;
	if (fullscreen)
		flags |= SDL_WINDOW_FULLSCREEN;
	else
		flags |= SDL_WINDOW_RESIZABLE;

	mainWindow = SDL_CreateWindow("SDL Template",
		SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED,
		width, height, flags);
	if (mainWindow == NULL)
	{
		cout << "SDL_Error() reports: " << SDL_GetError() << endl;
		ExitFailure("Unable to create window");
	}

	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 4);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 3);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_FLAGS, SDL_GL_CONTEXT_DEBUG_FLAG);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_COMPATIBILITY);
	
	// 4x antialiasing
	SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1);
	SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, 4);

	if (bpp == 24 || bpp == 32)
	{
		SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
		if (bpp == 24)
			SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 8);
		else if (bpp == 32)
			SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 0);
	}
	else if (bpp == 16)
	{
		SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
		SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 6);
		SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
	}
	else if (bpp == 15)
	{
		SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
		SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
		SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
	}
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
	SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);

	SDL_GL_SetSwapInterval(1);
	mainContext = SDL_GL_CreateContext(mainWindow);
	if (mainContext == NULL)
	{
		cout << "SDL_Error() reports: " << SDL_GetError() << endl;
		ExitFailure("Unable to create OpenGL Context");
	}

	glewExperimental = GL_TRUE;
	GLenum err = glewInit();
	if (err != GLEW_OK)
		ExitFailure("Unable to initialize GLEW\n");
	if(GLEW_ARB_debug_output)
	{
		glDebugMessageCallbackARB(&Fluxions::DebugMessageCallbackARB, NULL);
	}
	else if(GLEW_AMD_debug_output)
	{
		glDebugMessageCallbackAMD(&Fluxions::DebugMessageCallbackAMD, NULL);
	}


	char *versionString = const_cast<char *>((char*)glGetString(GL_VERSION));
	cout << "GL_VERSION: " << versionString << endl;
	versionString = const_cast<char *>((char*)glGetString(GL_VENDOR));
	cout << "GL_VENDOR:  " << versionString << endl;

	int img_init_flags = IMG_INIT_JPG | IMG_INIT_PNG;
	int img_initted = IMG_Init(img_init_flags);
	if (img_init_flags & img_initted != img_init_flags)
	{
		cout << "IMG_Init: Failed to init required jpg and png suppport.\n";
		cout << "IMG_GetError() reports: " << IMG_GetError() << endl;
	}

	OnResize(width, height);
}


void SDLTemplate::Quit()
{
	OnKill();
	IMG_Quit();
	SDL_GL_DeleteContext(mainContext);
	SDL_DestroyWindow(mainWindow);
	mainWindow = NULL;
	SDL_Quit();
}


void SDLTemplate::SendQuitEvent()
{
	SDL_Event event = {0};
	event.type = SDL_QUIT;
	SDL_PushEvent(&event);
}


void SDLTemplate::SwapBuffers()
{
	SDL_GL_SwapWindow(mainWindow);
}


void SDLTemplate::Init()
{
	OnInit();
}


void SDLTemplate::Kill()
{
	OnKill();
}

void SDLTemplate::Start()
{
	mainLoopRunning = true;
	mainLoopStarted = true;
	OnStart();
	MainEventLoop();
}


void SDLTemplate::MainEventLoop()
{
	int delta = 0;				
	SDL_Event event;
	while (mainLoopRunning)
	{
		while (SDL_PollEvent(&event))
		{
			switch (event.type)
			{
			case SDL_QUIT:
				Stop();
				break;
			case SDL_KEYDOWN:
				OnKeyDown(event.key.keysym.sym);
				break;
			case SDL_KEYUP:
				OnKeyUp(event.key.keysym.sym);
				break;
			case SDL_MOUSEBUTTONDOWN:
				OnMouseButtonDown(event.button.button);
				break;
			case SDL_MOUSEBUTTONUP:
				OnMouseButtonUp(event.button.button);
				break;
			case SDL_MOUSEMOTION:
				OnMouseMotion(event.motion.x, event.motion.y, event.motion.xrel, event.motion.yrel, event.motion.state);
				break;
			case SDL_JOYAXISMOTION:
				if (event.jaxis.which <= MAX_JOYSTICKS)
				{
					delta = event.jaxis.value - jaxis_value[event.jaxis.which];
					jaxis_value[event.jaxis.which] = event.jaxis.value;
				}
				OnJoyAxis(event.jaxis.axis, event.jaxis.value, delta);
				break;
			case SDL_JOYBALLMOTION:
				OnJoyBall(event.jball.which, event.jball.ball);
				break;
			case SDL_JOYBUTTONDOWN:
				OnJoyButtonDown(event.jbutton.button);
				break;
			case SDL_JOYBUTTONUP:
				OnJoyButtonUp(event.jbutton.button);
				break;
			case SDL_WINDOWEVENT:
				if (event.window.event == SDL_WINDOWEVENT_RESIZED)
					OnResize(event.window.data1, event.window.data2);
				break;
			default:
				break;
			}
		}
		t0 = t1;
		t1 = SDL_GetTicks() / 1000.0;
		double deltaTime = t1 - t0;
		OnUpdate(deltaTime);

		OnRender();
		OnRenderOverlay();
		SwapBuffers();
	}
}


void SDLTemplate::Stop()
{
	mainLoopRunning = false;
	while (SDL_PollEvent(NULL));
	OnStop();
}


void SDLTemplate::Pause()
{
	mainLoopRunning = false;
	while (SDL_PollEvent(NULL));
	OnPause();
}


void SDLTemplate::Resume()
{
	mainLoopRunning = true;
	OnResume();
	MainEventLoop();
}


void SDLTemplate::ExitFailure(const char *msg)
{
	printf("%s: %s\n", msg, SDL_GetError());
	SDL_Quit();
	exit(EXIT_FAILURE);
}


void SDLTemplate::RegisterApplication(Viperfish::Application *app)
{
	application = app;
}


void SDLTemplate::OnInit()
{
	if (application) application->OnInit();
}

void SDLTemplate::OnKill()
{
	if (application) application->OnKill();
}

void SDLTemplate::OnUpdate(double deltaTime)
{
	if (application) application->OnUpdate(deltaTime);
}

void SDLTemplate::OnTimer(int timerId, int value, double timerInterval)
{
	if (application) application->OnTimer(timerId, value, timerInterval);
}

void SDLTemplate::OnResize(int width, int height)
{
	windowWidth = width;
	windowHeight = height;
	if (application) application->OnResize(width, height);
}

void SDLTemplate::OnRender()
{
	// Set these, because they were probably screwed up last time...=D
	glViewport(0, 0, windowWidth, windowHeight);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	double fov = 45.0;
	double aspectRatio = (double)windowWidth / (double)windowHeight;
	double znear = 0.03;
	double zfar = 100.0;
	gluPerspective(fov, aspectRatio, znear, zfar);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	if (application) application->OnRender();
}

void SDLTemplate::OnRenderOverlay()
{
	// Set these, because they were probably screwed up last time...=D
	glViewport(0, 0, windowWidth, windowHeight);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, windowWidth, windowHeight, 0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	if (application) application->OnRenderOverlay();
}

void SDLTemplate::OnStart()
{
	if (application)
	{
		application->OnStart();
		application->OnResize(windowWidth, windowHeight);
	}
}

void SDLTemplate::OnStop()
{
	if (application) application->OnStop();
}

void SDLTemplate::OnPause()
{
	if (application) application->OnPause();
}

void SDLTemplate::OnResume()
{
	if (application) application->OnResume();
}

#ifndef GLUT_KEY_F1
#define  GLUT_KEY_F1                        0x0001
#define  GLUT_KEY_F2                        0x0002
#define  GLUT_KEY_F3                        0x0003
#define  GLUT_KEY_F4                        0x0004
#define  GLUT_KEY_F5                        0x0005
#define  GLUT_KEY_F6                        0x0006
#define  GLUT_KEY_F7                        0x0007
#define  GLUT_KEY_F8                        0x0008
#define  GLUT_KEY_F9                        0x0009
#define  GLUT_KEY_F10                       0x000A
#define  GLUT_KEY_F11                       0x000B
#define  GLUT_KEY_F12                       0x000C
#define  GLUT_KEY_LEFT                      0x0064
#define  GLUT_KEY_UP                        0x0065
#define  GLUT_KEY_RIGHT                     0x0066
#define  GLUT_KEY_DOWN                      0x0067
#define  GLUT_KEY_PAGE_UP                   0x0068
#define  GLUT_KEY_PAGE_DOWN                 0x0069
#define  GLUT_KEY_HOME                      0x006A
#define  GLUT_KEY_END                       0x006B
#define  GLUT_KEY_INSERT                    0x006C
#endif

void SDLTemplate::OnKeyUp(int key)
{
	if (application)
	{
		application->OnKeyUp(key);

		// GLUT compatibility
		if (key == SDLK_F1) application->OnSpecialKeyUp(GLUT_KEY_F1);
		if (key == SDLK_F2) application->OnSpecialKeyUp(GLUT_KEY_F2);
		if (key == SDLK_F3) application->OnSpecialKeyUp(GLUT_KEY_F3);
		if (key == SDLK_F4) application->OnSpecialKeyUp(GLUT_KEY_F4);
		if (key == SDLK_F5) application->OnSpecialKeyUp(GLUT_KEY_F5);
		if (key == SDLK_F6) application->OnSpecialKeyUp(GLUT_KEY_F6);
		if (key == SDLK_F7) application->OnSpecialKeyUp(GLUT_KEY_F7);
		if (key == SDLK_F8) application->OnSpecialKeyUp(GLUT_KEY_F8);
		if (key == SDLK_F9) application->OnSpecialKeyUp(GLUT_KEY_F9);
		if (key == SDLK_F10) application->OnSpecialKeyUp(GLUT_KEY_F10);
		if (key == SDLK_F11) application->OnSpecialKeyUp(GLUT_KEY_F11);
		if (key == SDLK_F12) application->OnSpecialKeyUp(GLUT_KEY_F12);
		if (key == SDLK_LEFT) application->OnSpecialKeyUp(GLUT_KEY_LEFT);
		if (key == SDLK_RIGHT) application->OnSpecialKeyUp(GLUT_KEY_RIGHT);
		if (key == SDLK_UP) application->OnSpecialKeyUp(GLUT_KEY_UP);
		if (key == SDLK_DOWN) application->OnSpecialKeyUp(GLUT_KEY_DOWN);
		if (key == SDLK_PAGEUP) application->OnSpecialKeyUp(GLUT_KEY_PAGE_UP);
		if (key == SDLK_PAGEDOWN) application->OnSpecialKeyUp(GLUT_KEY_PAGE_DOWN);
		if (key == SDLK_HOME) application->OnSpecialKeyUp(GLUT_KEY_HOME);
		if (key == SDLK_END) application->OnSpecialKeyUp(GLUT_KEY_END);
		if (key == SDLK_INSERT) application->OnSpecialKeyUp(GLUT_KEY_INSERT);
	}
}

void SDLTemplate::OnKeyDown(int key)
{
	if (application)
	{
		application->OnKeyDown(key);
		// GLUT compatibility
		if (key == SDLK_F1) application->OnSpecialKeyDown(GLUT_KEY_F1);
		if (key == SDLK_F2) application->OnSpecialKeyDown(GLUT_KEY_F2);
		if (key == SDLK_F3) application->OnSpecialKeyDown(GLUT_KEY_F3);
		if (key == SDLK_F4) application->OnSpecialKeyDown(GLUT_KEY_F4);
		if (key == SDLK_F5) application->OnSpecialKeyDown(GLUT_KEY_F5);
		if (key == SDLK_F6) application->OnSpecialKeyDown(GLUT_KEY_F6);
		if (key == SDLK_F7) application->OnSpecialKeyDown(GLUT_KEY_F7);
		if (key == SDLK_F8) application->OnSpecialKeyDown(GLUT_KEY_F8);
		if (key == SDLK_F9) application->OnSpecialKeyDown(GLUT_KEY_F9);
		if (key == SDLK_F10) application->OnSpecialKeyDown(GLUT_KEY_F10);
		if (key == SDLK_F11) application->OnSpecialKeyDown(GLUT_KEY_F11);
		if (key == SDLK_F12) application->OnSpecialKeyDown(GLUT_KEY_F12);
		if (key == SDLK_LEFT) application->OnSpecialKeyDown(GLUT_KEY_LEFT);
		if (key == SDLK_RIGHT) application->OnSpecialKeyDown(GLUT_KEY_RIGHT);
		if (key == SDLK_UP) application->OnSpecialKeyDown(GLUT_KEY_UP);
		if (key == SDLK_DOWN) application->OnSpecialKeyDown(GLUT_KEY_DOWN);
		if (key == SDLK_PAGEUP) application->OnSpecialKeyDown(GLUT_KEY_PAGE_UP);
		if (key == SDLK_PAGEDOWN) application->OnSpecialKeyDown(GLUT_KEY_PAGE_DOWN);
		if (key == SDLK_HOME) application->OnSpecialKeyDown(GLUT_KEY_HOME);
		if (key == SDLK_END) application->OnSpecialKeyDown(GLUT_KEY_END);
		if (key == SDLK_INSERT) application->OnSpecialKeyDown(GLUT_KEY_INSERT);
	}
}

void SDLTemplate::OnSpecialKeyDown(int key)
{
	if (application) application->OnSpecialKeyDown(key);
}

void SDLTemplate::OnSpecialKeyUp(int key)
{
	if (application) application->OnSpecialKeyUp(key);
}

void SDLTemplate::OnMouseButtonDown(int button)
{
	if (application) application->OnMouseButtonDown(button);
}

void SDLTemplate::OnMouseButtonUp(int button)
{
	if (application) application->OnMouseButtonUp(button);
}

void SDLTemplate::OnMouseMotion(int x, int y, int xrel, int yrel, int buttons)
{
	if (application) application->OnMouseMotion(x, y, xrel, yrel, buttons);
}

void SDLTemplate::OnJoyAxis(int axis, float value, float delta)
{
	if (application) application->OnJoyAxis(axis, value, delta);
}

void SDLTemplate::OnJoyHat(int direction)
{
	if (application) application->OnJoyHat(direction);
}

void SDLTemplate::OnJoyBall(int ball, float value)
{
	if (application) application->OnJoyBall(ball, value);
}

void SDLTemplate::OnJoyButtonDown(int button)
{
	if (application) application->OnJoyButtonDown(button);
}

void SDLTemplate::OnJoyButtonUp(int button)
{
	if (application) application->OnJoyButtonUp(button);
}


SDLTemplate SDL;
