// Fluxions 3D Engine
// simple_rendertotexture.cpp
// Copyright (C) 2000-2011 Jonathan Metzgar
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

#include <iostream>
#include <fluxions/simple_rendertotexture.hpp>

namespace Fluxions
{
	SimpleRenderToTexture::SimpleRenderToTexture()
	{
		framebuffer = 0;
		colorBuffer = 0;
		depthBuffer = 0;
		depthStencilBuffer = 0;
		colorTexture = 0;
		depthTexture = 0;
		depthStencilTexture = 0;
		savedViewport[0] = 0;
		savedViewport[1] = 0;
		savedViewport[2] = 0;
		savedViewport[3] = 0;
		viewport[0] = 0;
		viewport[1] = 0;
		viewport[2] = 0;
		viewport[3] = 0;
		framebufferWidth = 0;
		framebufferHeight = 0;
		usingColorBuffer = GL_FALSE;
		usingDepthBuffer = GL_FALSE;
		usingDepthStencilBuffer = GL_FALSE;
	}

	void SimpleRenderToTexture::Reset()
	{
		savedViewport[0] = 0;
		savedViewport[1] = 0;
		savedViewport[2] = 0;
		savedViewport[3] = 0;
		viewport[0] = 0;
		viewport[1] = 0;
		viewport[2] = 0;
		viewport[3] = 0;
		framebufferWidth = 0;
		framebufferHeight = 0;
		usingColorBuffer = GL_FALSE;
		usingDepthBuffer = GL_FALSE;
		usingDepthStencilBuffer = GL_FALSE;

		if (framebuffer)
			glDeleteFramebuffers(1, &framebuffer);
		if (colorBuffer)
			glDeleteRenderbuffers(1, &colorBuffer);
		if (depthBuffer)
			glDeleteRenderbuffers(1, &depthBuffer);
		if (depthStencilBuffer)
			glDeleteRenderbuffers(1, &depthStencilBuffer);
		if (colorTexture)
			glDeleteTextures(1, &colorTexture);
		if (depthTexture)
			glDeleteTextures(1, &depthTexture);
		if (depthStencilTexture)
			glDeleteTextures(1, &depthStencilTexture);
		framebuffer = 0;
		colorBuffer = 0;
		depthBuffer = 0;
		depthStencilBuffer = 0;
		colorTexture = 0;
		depthTexture = 0;
		depthStencilTexture = 0;
	}

	void SimpleRenderToTexture::Init(GLenum type, GLbitfield buffers, int width, int height)
	{
		textureType = type;

		if (buffers & GL_COLOR_BUFFER_BIT) usingColorBuffer = GL_TRUE;
		if (buffers & GL_DEPTH_BUFFER_BIT) usingDepthBuffer = GL_TRUE;
		if (buffers & GL_STENCIL_BUFFER_BIT) usingDepthStencilBuffer = GL_TRUE;

		// Get framebuffers and renderbuffers from OpenGL
		if (!framebuffer)
		{
			glGenFramebuffers(1, &framebuffer);
			glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);
		}

		if (usingColorBuffer)
		{
			//if (!colorBuffer) glGenRenderbuffers(1, &colorBuffer);
			if (!colorTexture) glGenTextures(1, &colorTexture);
			//glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_RENDERBUFFER, colorBuffer);
		}
		if (usingDepthBuffer)
		{
			if (!depthBuffer)
				glGenRenderbuffers(1, &depthBuffer);
			if (!depthTexture)
				glGenTextures(1, &depthTexture);
			glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, depthBuffer);
		}
		if (usingDepthStencilBuffer)
		{
			if (!depthStencilBuffer)
				glGenRenderbuffers(1, &depthStencilBuffer);
			if (!depthStencilTexture)
				glGenTextures(1, &depthStencilTexture);
			glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_STENCIL_ATTACHMENT, GL_RENDERBUFFER, depthStencilBuffer);
		}

		Resize(width, height);

		glBindFramebuffer(GL_FRAMEBUFFER, 0);
	}

	void SimpleRenderToTexture::Resize(int width, int height)
	{
		viewport[0] = 0;
		viewport[1] = 0;
		viewport[2] = width;
		viewport[3] = height;
		framebufferWidth = width;
		framebufferHeight = height;

		if (usingColorBuffer)
		{
			//glBindRenderbuffer(GL_RENDERBUFFER, colorBuffer);
			//glRenderbufferStorage(GL_RENDERBUFFER, GL_COLOR_ATTACHMENT0, framebufferWidth, framebufferHeight);
			glBindTexture(textureType, colorTexture);
			glTexParameteri(textureType, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			glTexParameteri(textureType, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			glTexParameteri(textureType, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
			glTexParameteri(textureType, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
			if (textureType == GL_TEXTURE_CUBE_MAP)
			{
				for (int i = 0; i < 6; i++)
				{
					glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
				}
				glBindTexture(textureType, 0);
				glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, colorTexture, 0);
			}
			else
			{
				glTexImage2D(textureType, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
				glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, textureType, colorTexture, 0);
			}
		}
		if (usingDepthBuffer)
		{
			glBindRenderbuffer(GL_RENDERBUFFER, depthBuffer);
			glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, framebufferWidth, framebufferHeight);
			glBindTexture(textureType, depthTexture);
			glTexParameteri(textureType, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			glTexParameteri(textureType, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			glTexParameteri(textureType, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
			glTexParameteri(textureType, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
			if (textureType == GL_TEXTURE_CUBE_MAP)
			{
				for (int i = 0; i < 6; i++)
				{
					glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_DEPTH_COMPONENT, width, height, 0, GL_DEPTH_COMPONENT, GL_FLOAT, 0);
					glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, depthTexture, 0);
				}
			}
			else
			{
				glTexImage2D(textureType, 0, GL_DEPTH_COMPONENT, width, height, 0, GL_DEPTH_COMPONENT, GL_FLOAT, 0);
				glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, textureType, depthTexture, 0);
			}			
		}
		if (usingDepthStencilBuffer)
		{
			glBindRenderbuffer(GL_RENDERBUFFER, depthStencilBuffer);
			glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH24_STENCIL8, framebufferWidth, framebufferHeight);
			glBindTexture(textureType, depthStencilTexture);
			glTexParameteri(textureType, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			glTexParameteri(textureType, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			glTexParameteri(textureType, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
			glTexParameteri(textureType, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
			if (textureType == GL_TEXTURE_CUBE_MAP)
			{
				for (int i = 0; i < 6; i++)
				{
					glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X + i, 0, GL_DEPTH24_STENCIL8, width, height, 0, GL_DEPTH24_STENCIL8, GL_UNSIGNED_INT_24_8, 0);
					glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH24_STENCIL8, GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, depthStencilTexture, 0);
				}
			}
			else
			{
				glTexImage2D(textureType, 0, GL_DEPTH24_STENCIL8, width, height, 0, GL_DEPTH24_STENCIL8, GL_UNSIGNED_INT_24_8, 0);
				glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH24_STENCIL8, textureType, depthStencilTexture, 0);
			}			
		}

		glBindTexture(textureType, 0);

		GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
		if (status != GL_FRAMEBUFFER_COMPLETE)
		{
			std::cout << "class SimpleRenderToTexture: ";
			if (status == GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT)
				std::cout << "GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT\n";
			if (status == GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER)
				std::cout << "GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER\n";
			if (status == GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS)
				std::cout << "GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS\n";
			if (status == GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT)
				std::cout << "GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT\n";
			if (status == GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE)
				std::cout << "GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE\n";
			if (status == GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER)
				std::cout << "GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER\n";
		}
		else
		{
			std::cout << "class SimpleRenderToTexture: ";
			std::cout << "Framebuffer is complete!\n";
		}
	}

	void SimpleRenderToTexture::SetViewport(int x, int y, int width, int height)
	{
		viewport[0] = x;
		viewport[1] = y;
		viewport[2] = width;
		viewport[3] = height;
	}

	void SimpleRenderToTexture::Clear(GLbitfield buffers)
	{
		GLbitfield buffersToClear = 0;
		if (buffers == 0)
		{
			if (usingColorBuffer)
				buffersToClear |= GL_COLOR_BUFFER_BIT;
			if (usingDepthBuffer)
				buffersToClear |= GL_DEPTH_BUFFER_BIT;
			if (usingDepthStencilBuffer)
				buffersToClear |= GL_STENCIL_BUFFER_BIT;
		}
		glClear(buffersToClear);
	}

	void SimpleRenderToTexture::BindCubeMapTexture(GLenum activeTexture, GLenum face, GLbitfield mask)
	{
		glActiveTexture(activeTexture);
		if (mask == GL_COLOR_BUFFER_BIT)
		{
			glBindTexture(face, colorTexture);
		}
		if (mask == GL_DEPTH_BUFFER_BIT)
		{
			glBindTexture(face, depthTexture);
		}
		if (mask == GL_STENCIL_BUFFER_BIT)
		{
			glBindTexture(face, depthStencilTexture);
		}
		if (mask == 0)
			glBindTexture(face, 0);
		glActiveTexture(GL_TEXTURE0);
	}

	void SimpleRenderToTexture::BindTexture(GLenum activeTexture, GLbitfield mask)
	{
		glActiveTexture(activeTexture);
		if (mask == GL_COLOR_BUFFER_BIT)
		{
			glBindTexture(textureType, colorTexture);
		}
		if (mask == GL_DEPTH_BUFFER_BIT)
		{
			glBindTexture(textureType, depthTexture);
		}
		if (mask == GL_STENCIL_BUFFER_BIT)
		{
			glBindTexture(textureType, depthStencilTexture);
		}
		if (mask == 0)
			glBindTexture(textureType, 0);
		glActiveTexture(GL_TEXTURE0);
	}

	void SimpleRenderToTexture::BeginScene()
	{
		glGetIntegerv(GL_VIEWPORT, savedViewport);
		glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);
		glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);
	}

	void SimpleRenderToTexture::EndScene()
	{
		glViewport(savedViewport[0], savedViewport[1], savedViewport[2], savedViewport[3]);
		glBindFramebuffer(GL_FRAMEBUFFER, 0);
	}
}
