#include <viperfish/visualization.hpp>
#include <fluxions/ftl_math.hpp>
#include <fluxions/glut_extensions.hpp>
#include <fluxions/html_colors.hpp>

namespace Viperfish
{
	using namespace Fluxions;

	VisualizationApp::VisualizationApp()
	{
		for (int i = 0; i < 256; i++)
			keys[i] = 0;
		mouseX = mouseY = mouseLastX = mouseLastY = mouseDx = mouseDy = 0;
		mouseButtons = mouseLastButtons = 0;
		screenWidth = 0;
		screenHeight = 0;
		znear = 0.3;
		zfar = 100.0;
		fovy = 45.0;
		screenAspect = 0;
		elapsedTime = 0;
		framecount = 0;
		renderTime = 0;

		applicationTitle = "Application Title";
		copyrightStatement = "Copyright Statement";
	}

	VisualizationApp::~VisualizationApp()
	{
	}

	void VisualizationApp::OnStart()
	{
		elapsedTime = 0;
		renderTime = 0;
		framecount = 0;
	}

	void VisualizationApp::OnKeyDown(int key)
	{
		if (key >= 128)
			return;
		keys[key] = 1;
	}

	void VisualizationApp::OnKeyUp(int key)
	{
		if (key >= 128)
			return;
		keys[key] = 0;
	}

	void VisualizationApp::OnSpecialKeyDown(int key)
	{
		if (key >= 128)
			return;
		keys[key + 128] = 1;
	}

	void VisualizationApp::OnSpecialKeyUp(int key)
	{
		if (key >= 128)
			return;
		keys[key + 128] = 0;
	}

	void VisualizationApp::OnMouseMotion(int x, int y, int xrel, int yrel, int buttons)
	{
		mouseLastX = mouseX;
		mouseLastY = mouseY;
		mouseX = x;
		mouseY = y;
		mouseDx = mouseX - mouseLastX;
		mouseDy = mouseY - mouseLastY;
		mouseLastButtons = mouseButtons;
		mouseButtons = buttons;
	}

	void VisualizationApp::OnMouseButtonUp(int button)
	{
		unsigned mask = ~(1<<(button-1));
		mouseButtons &= mask;
	}

	void VisualizationApp::OnMouseButtonDown(int button)
	{
		unsigned mask = (1<<(button-1));
		mouseButtons |= mask;
	}

	void VisualizationApp::OnResize(int width, int height)
	{
		screenWidth = width;
		screenHeight = height;
		screenAspect = (double)width / (double)height;
		projMatrix.LoadIdentity();
		projMatrix.Perspective(fovy, screenAspect, znear, zfar);
		glMatrixMode(GL_PROJECTION);
		glLoadMatrixf(projMatrix.m);
		glMatrixMode(GL_MODELVIEW);
		glViewport(0, 0, width, height);
	}

	void VisualizationApp::OnUpdate(double deltaTime)
	{
		elapsedTime += deltaTime;
		renderTime += deltaTime;
		dt = deltaTime;

		if (keys[27])
			exit(EXIT_SUCCESS);

		float xrot, yrot, zrot, xt, yt, zt, dt;
		xrot = yrot = zrot = 0;
		xt = yt = zt = 0;
		dt = deltaTime;

		if (keys[GLUT_KEY_LEFT + 128]) yrot = 30;
		if (keys[GLUT_KEY_RIGHT + 128]) yrot = -30;
		if (keys[GLUT_KEY_UP + 128]) xrot = -30;
		if (keys[GLUT_KEY_DOWN + 128]) xrot = 30;
		if (keys['q'] || keys['Q']) zrot = 30;
		if (keys['e'] || keys['E']) zrot = -30;
		if (keys['w'] || keys['W']) zt = -1;
		if (keys['s'] || keys['S']) zt = 1;
		if (keys['a'] || keys['A']) xt = -1;
		if (keys['d'] || keys['D']) xt = 1;
		if (keys['z'] || keys['Z']) yt = 1;
		if (keys['c'] || keys['C']) yt = -1;

		if (mouseButtons)
		{
			xrot = 10 * -mouseDy;
			yrot = 10 * -mouseDx;
		}

		if (currentObjectMatrix != "")
		{
			//objectMatrix[currentObjectMatrix].Rotate(xrot * dt, ftl::Vector3f(1.0f, 0.0f, 0.0f));
			//objectMatrix[currentObjectMatrix].Rotate(yrot * dt, ftl::Vector3f(0.0f, 1.0f, 0.0f));
			//objectMatrix[currentObjectMatrix].Rotate(zrot * dt, ftl::Vector3f(0.0f, 0.0f, 1.0f));
			objectMatrix[currentObjectMatrix].Translate(xt * dt, yt * dt, zt * dt);
		}
		else
		{
			cameraMatrix.Rotate(xrot * dt, ftl::Vector3f(1, 0, 0));
			cameraMatrix.Rotate(yrot * dt, ftl::Vector3f(0, 1, 0));
			cameraMatrix.Rotate(zrot * dt, ftl::Vector3f(0, 0, 1));
			cameraMatrix.Translate(xt * dt, yt * dt, zt * dt);
		}
	}

	void VisualizationApp::OnRender()
	{
		projMatrix.LoadIdentity();
		projMatrix.Perspective(fovy, screenAspect, znear, zfar);
		glMatrixMode(GL_PROJECTION);
		glLoadMatrixf(projMatrix.m);
		glMatrixMode(GL_MODELVIEW);
	}

	void VisualizationApp::OnRenderOverlay()
	{
		framecount++;
		fps = framecount / renderTime;
		if (framecount > 240)
		{
			framecount = 0;
			renderTime = 0;
		}

		glColor4fv(Colors::Lime);
		
		int y = 15;
		glutBitmapPrintLeft(GLUT_BITMAP_9_BY_15, y, applicationTitle.c_str());
		y += 15;
		glutBitmapPrintLeft(GLUT_BITMAP_9_BY_15, y, copyrightStatement.c_str());
		y += 30;

		y = 15;
		glutBitmapPrintRight(GLUT_BITMAP_9_BY_15, y, "%3.2f fps", floor(fps-0.5));

		glColor3d(1, 1, 1);
	}
}
