/*
GLUT Template by Jonathann Metzgar
Copyright (C) 2013 Jonathan Metzgar

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#define _STDCALL_SUPPORTED
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <GL/glut.h>

#define _USE_MATH_DEFINES
#include <math.h>

double g_distance = 0;

char *g_windowTitle = "GLUT Template";
int g_displayMode = GLUT_DOUBLE | GLUT_STENCIL | GLUT_RGBA | GLUT_DEPTH;
double g_Fps = 0;
double g_aspectRatio = 1.333;
int g_screenWidth = 640;
int g_screenHeight = 480;
int g_mouseX = 0;
int g_mouseY = 0;
int g_mouseButtons = 0;
int g_joystickButtonMask = 0;
int g_joystickX = 0;
int g_joystickY = 0;
int g_joystickZ = 0;
int g_joystickPollInterval = 16;
int g_timer1millis = 0;
int g_timer2millis = 0;
int g_timer3millis = 0;
int g_timer4millis = 0;
int g_timer1value = 0;
int g_timer2value = 0;
int g_timer3value = 0;
int g_timer4value = 0;
double g_updateInterval = 0.01666667f;
double g_renderInterval = 0.01666667f;
double g_Fov = 45.0;
double g_zNear = 0.01;
double g_zFar = 100.0;
bool g_keyMap[256] = {false};
bool g_specialKeyMap[256] = {false};

void OnInit();
void OnUpdate(double deltaTime);
void OnRender();
void OnRenderOverlay();

void Set2DViewport();
void Set3DViewport();
void glutBitmapString(void *font, const char *string);
void glutStrokeString(void *font, const char *string);

void glutSolidSphericalFunction(float radius, int slices, int stacks, float (*f)(float theta, float phi));

void display();
void reshape(int width, int height);
void keyboard(unsigned char key, int x, int y);
void keyboardup(unsigned char key, int x, int y);
void special(int key, int x, int y);
void specialup(int key, int x, int y);
void mouse(int button, int state, int x, int y);
void motion(int x, int y);
void passive(int x, int y);
void entry(int state);
void visibility(int state);
void idle();
void timer1(int value);
void timer2(int value);
void timer3(int value);
void timer4(int value);
void menu(int state);
void menustatus(int state, int x, int y);
void windowstatus(int state);
void joystick(unsigned int buttonMask, int x, int y, int z);

float constantRadius(float theta, float phi)
{
	return 2+cos(2*theta);
}

void glutSolidSphericalFunction(float radius, int slices, int stacks, float (*f)(float theta, float phi))
{
	float u = 0.0;
	float v = 0.0;
	float du = 2*M_PI / slices;
	float dv = M_PI / stacks;

	glNormal3f(0, 0, 1);

	glBegin(GL_POINTS);
	for (int i = 0; i < slices; i++)
	{
		for (int j = 0; j < stacks; j++)
		{
			float r00 = radius * f(u, v);
			float r01 = radius * f(u + du, v);
			float r10 = radius * f(u + du, v + dv);
			float r11 = radius * f(u, v + dv);
			float nx = cos(u) * sin(v);
			float ny = sin(u) * sin(v);
			float nz = cos(v);
			float x = r00 * nx;
			float y = r00 * ny;
			float z = r00 * nz;
			glNormal3f(nx, ny, nz);
			glVertex3f(x, y, z);
			v += dv;
		}
		v = 0.0;
		u += du;
	}
	glEnd();
}


int main(int argc, char **argv)
{
	printf("GLUT Template\nCopyright (C) 2013 Jonathan Metzgar\n\n");
	printf("This program is free software: you can redistribute it and/or modify\nit under the terms of the GNU General Public License as published by\nthe Free Software Foundation, either version 3 of the License, or\n(at your option) any later version.\n\n");
	printf("This program is distributed in the hope that it will be useful,\nbut WITHOUT ANY WARRANTY; without even the implied warranty of\nMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\nGNU General Public License for more details.\n\n");
	printf("You should have received a copy of the GNU General Public License\nalong with this program.  If not, see <http://www.gnu.org/licenses/>.\n\n");

	glutInit(&argc, argv);
	glutInitDisplayMode(g_displayMode);
	glutInitWindowSize(g_screenWidth, g_screenHeight);
	glutCreateWindow(g_windowTitle);
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyboardup);
	glutSpecialFunc(special);
	glutSpecialUpFunc(specialup);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutPassiveMotionFunc(passive);
	glutEntryFunc(entry);
	glutVisibilityFunc(visibility);
	glutIdleFunc(idle);
	if (g_timer1millis != 0) glutTimerFunc(g_timer1millis, timer1, g_timer1value);
	if (g_timer2millis != 0) glutTimerFunc(g_timer2millis, timer2, g_timer2value);
	if (g_timer3millis != 0) glutTimerFunc(g_timer3millis, timer3, g_timer3value);
	if (g_timer4millis != 0) glutTimerFunc(g_timer4millis, timer4, g_timer4value);
	glutMenuStateFunc(menu);
	glutMenuStatusFunc(menustatus);
	glutWindowStatusFunc(windowstatus);
	glutJoystickFunc(joystick, g_joystickPollInterval);
	OnInit();
	glutMainLoop();
	return 0;
}

void OnInit()
{

}

float xrot = 0;

void OnUpdate(double deltaTime)
{
	xrot += 35 * deltaTime;
}

void RenderCheckerboard()
{
	for (int i = -4; i < 4; i++)
	{
		for (int j = -4; j < 4; j++)
		{
			glNormal3f(0, 1, 0);
			if ((j % 2 && i % 2) || j == i || (!(j % 2) && !(i % 2)))
				glColor3f(.7f, .7f, .7f);
			else
				glColor3f(.3f, .3f, .3f);
			glRectf(2 * i, 2 * j, 2*(i+1), 2*(j+1));
		}
	}
}

void OnRender()
{
	glColor3f(0.8f, 0.3f, 0.2f);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_COLOR_MATERIAL);


	glLoadIdentity();
	glTranslatef(0, 0, -20 - g_distance);
	//glutSolidTeapot(2.0);
	glPushMatrix();
	glRotatef(xrot, 0, 1, 0);
	glutSolidSphericalFunction(2.0, 128, 128, constantRadius);
	glPopMatrix();
	glColor3f(1, 0, 0);
	glLoadIdentity();
	glTranslatef(0, -4, -20 - g_distance);
	glRotatef(-90, 1, 0, 0);
	glRectf(-1, -1, 1, 1);

	RenderCheckerboard();

	glDisable(GL_COLOR_MATERIAL);
	glDisable(GL_LIGHT0);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
}

void OnRenderOverlay()
{
	glColor3f(1, 1, 1);
	
	int y = 20;
	int width;
	char fpsString[45];
	sprintf(fpsString, "GLUT Template", g_Fps);
	width = glutBitmapLength(GLUT_BITMAP_9_BY_15, (unsigned char *)fpsString);
	glRasterPos2f(g_screenWidth - width, y);
	glutBitmapString(GLUT_BITMAP_9_BY_15, fpsString);
	y += 15;

	sprintf(fpsString, "Jonathan Metzgar", g_Fps);
	width = glutBitmapLength(GLUT_BITMAP_9_BY_15, (unsigned char *)fpsString);
	glRasterPos2f(g_screenWidth - width, y);
	glutBitmapString(GLUT_BITMAP_9_BY_15, fpsString);
	y += 30;

	sprintf(fpsString, "%3.3f fps", g_Fps);
	width = glutBitmapLength(GLUT_BITMAP_9_BY_15, (unsigned char *)fpsString);
	glRasterPos2f(g_screenWidth - width, y);
	glutBitmapString(GLUT_BITMAP_9_BY_15, fpsString);
	y += 15;

	sprintf(fpsString, "-/=......ZOOM IN/OUT", g_Fps);
	width = glutBitmapLength(GLUT_BITMAP_9_BY_15, (unsigned char *)fpsString);
	glRasterPos2f(g_screenWidth - width, y);
	glutBitmapString(GLUT_BITMAP_9_BY_15, fpsString);
	y += 15;

	sprintf(fpsString, "ESC.............QUIT", g_Fps);
	width = glutBitmapLength(GLUT_BITMAP_9_BY_15, (unsigned char *)fpsString);
	glRasterPos2f(g_screenWidth - width, y);
	glutBitmapString(GLUT_BITMAP_9_BY_15, fpsString);
	y += 15;



	// Draw Robot Information
	y = 20;
	glColor3f(1, 0, 1);
	glRasterPos2f(0, y);
	sprintf(fpsString, "Zoom: %3f", g_distance);
	glutBitmapString(GLUT_BITMAP_9_BY_15, fpsString);
	y+=15;
}

void Set2DViewport()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, g_screenWidth, g_screenHeight, 0);
	glMatrixMode(GL_MODELVIEW);
}

void Set3DViewport()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(g_Fov, g_aspectRatio, g_zNear, g_zFar);
	glMatrixMode(GL_MODELVIEW);
}

void glutBitmapString(void *font, const char *string)
{
	int len = strlen(string);
	int i = 0;
	
	for (i = 0; i < len; i++)
	{
		glutBitmapCharacter(font, string[i]);
	}	
}

void glutStrokeString(void *font, const char *string)
{
	int len = strlen(string);
	int i = 0;
	
	for (i = 0; i < len; i++)
	{
		glutStrokeCharacter(font, string[i]);
	}	
}

void display()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	glClearColor(0.0, 0.0, 0.0, 1.0);
	
	glLoadIdentity();
	Set3DViewport();
	OnRender();
	
	glLoadIdentity();
	Set2DViewport();
	OnRenderOverlay();
	
	if (g_displayMode & GLUT_DOUBLE)
	{
		glutSwapBuffers();
	}
}

void reshape(int width, int height)
{
	g_aspectRatio = (double)width / (double)height;
	g_screenWidth = width;
	g_screenHeight = height;
	glViewport(0, 0, width, height);
}

void keyboard(unsigned char key, int x, int y)
{
	g_keyMap[key] = true;
	
	if (key == 27)
		exit(0);
	if (key == '=')
		g_distance += 1;
	if (key == '-')
		g_distance -= 1;
}

void keyboardup(unsigned char key, int x, int y)
{
	g_keyMap[key] = false;
}

void special(int key, int x, int y)
{
	g_specialKeyMap[key] = true;
}

void specialup(int key, int x, int y)
{
	g_specialKeyMap[key] = false;
}

void mouse(int button, int state, int x, int y)
{
	int buttonMask = 1 << button;
	if (state == GLUT_UP)
	{
		g_mouseButtons &= ~buttonMask;
	}
	else
	{
		g_mouseButtons |= buttonMask;
	}
	g_mouseX = x;
	g_mouseY = y;
}

void motion(int x, int y)
{
	g_mouseX = x;
	g_mouseY = y;
}

void passive(int x, int y)
{
	g_mouseX = x;
	g_mouseY = y;
}

void entry(int state)
{
}

void visibility(int state)
{
}

void idle()
{
	static double time0 = 0, time1 = 0;
	static double renderdt = 0;
	static double updatedt = 0;
	static int fps_count = 0;
	static double fps0 = 0, fps1 = 0;
	
	time0 = time1;
	time1 = glutGet(GLUT_ELAPSED_TIME) / 1000.0;
	renderdt += time1 - time0;
	updatedt += time1 - time0;
	
	if (fps_count == 20)
	{
		fps0 = fps1;
		fps1 = time1;
		fps_count = 0;
	}
	
	g_Fps = 20.0 / (fps1 - fps0);
	
	if (updatedt > g_updateInterval)
	{
		OnUpdate(updatedt);
		updatedt = 0;
	}
	
	if (renderdt > g_renderInterval)
	{
		fps_count++;	
		glutPostRedisplay();
		renderdt = 0;
	}
}

void timer1(int value)
{
}

void timer2(int value)
{
}

void timer3(int value)
{
}

void timer4(int value)
{
}

void menu(int state)
{
}

void menustatus(int state, int x, int y)
{
}

void windowstatus(int state)
{
}

void joystick(unsigned int buttonMask, int x, int y, int z)
{
	g_joystickButtonMask = buttonMask;
	g_joystickX = x;
	g_joystickY = y;
	g_joystickZ = z;
}
