#include "stdafx.h"
#include "xor_allegro.h"
#include "xor_sound.h"


AllegroGameEngine age;


AllegroGameEngine::AllegroGameEngine()
{
	setName("AllegroGameEngine");
}


AllegroGameEngine::~AllegroGameEngine()
{

}


int AllegroGameEngine::Init(int width, int height)
{
	actLikeContainer = false;

	if (isInitted)
		Kill();

	if (!al_init())
	{
		log("failed to initialize allegro!");
		return -1;
	}

	al_install_keyboard();
	al_install_joystick();
	al_install_mouse();
	al_init_font_addon();
	al_init_ttf_addon();
	al_init_image_addon();
	al_init_primitives_addon();

	SS.Init();

	display = al_create_display(width, height);
	if (!display)
	{
		log("failed to create display!");
		return -1;
	}

	eventQueue = al_create_event_queue();
	if (!eventQueue)
	{
		log("failed to create event_queue!");
		al_destroy_display(display);
		return -1;
	}

	al_register_event_source(eventQueue, al_get_display_event_source(display));
	al_register_event_source(eventQueue, al_get_keyboard_event_source());
	al_register_event_source(eventQueue, al_get_mouse_event_source());
	al_register_event_source(eventQueue, al_get_joystick_event_source());

	if (!(fonts["PetMe8"] = al_load_ttf_font("data/PetMe.ttf", -8, 0))) log("PetMe8 could not be loaded.");
	if (!(fonts["PetMe16"] = al_load_ttf_font("data/PetMe.ttf", -16, 0))) log("PetMe16 could not be loaded.");
	if (!(fonts["PetMe24"] = al_load_ttf_font("data/PetMe.ttf", -24, 0))) log("PetMe24 could not be loaded.");
	if (!(fonts["PetMe32"] = al_load_ttf_font("data/PetMe.ttf", -32, 0))) log("PetMe24 could not be loaded.");

	return 0;
}


void AllegroGameEngine::Kill()
{
	isInitted = false;
	al_destroy_display(display);
	al_destroy_event_queue(eventQueue);
	SS.Kill();
}


void AllegroGameEngine::Start()
{
	isRunning = true;
	OnStart();
}


void AllegroGameEngine::Stop()
{
	isRunning = false;
	OnStop();
}


bool AllegroGameEngine::MainLoopIteration()
{
	if (!isRunning)
		return false;

	// Updated Timer
	t0 = t1;
	t1 = al_get_time();
	dt = t1 - t0;
	elapsedTime += dt;

	// Handle Events
	ALLEGRO_EVENT ev;

	while (al_get_next_event(eventQueue, &ev))
	{
		switch (ev.type)
		{
		case ALLEGRO_EVENT_JOYSTICK_AXIS:
			OnJoystickAxis(ev.joystick.id, ev.joystick.stick, ev.joystick.axis, ev.joystick.pos);
			break;
		case ALLEGRO_EVENT_JOYSTICK_BUTTON_DOWN:
			OnJoystickButtonDown(ev.joystick.id, ev.joystick.button);
			break;
		case ALLEGRO_EVENT_JOYSTICK_BUTTON_UP:
			OnJoystickButtonUp(ev.joystick.id, ev.joystick.button);
			break;
		case ALLEGRO_EVENT_JOYSTICK_CONFIGURATION:
			OnJoystickConfiguration();
			break;
		case ALLEGRO_EVENT_KEY_DOWN:
			OnKeyDown(ev.keyboard.keycode, ev.keyboard.display);
			break;
		case ALLEGRO_EVENT_KEY_UP:
			OnKeyUp(ev.keyboard.keycode, ev.keyboard.display);
			break;
		case ALLEGRO_EVENT_KEY_CHAR:
			OnKeyChar(ev.keyboard.keycode, ev.keyboard.unichar, ev.keyboard.modifiers, ev.keyboard.repeat, ev.keyboard.display);
			break;
		case ALLEGRO_EVENT_MOUSE_AXES:
			OnMouseAxes(ev.mouse.x, ev.mouse.y, ev.mouse.z, ev.mouse.w, ev.mouse.dx, ev.mouse.dy, ev.mouse.dz, ev.mouse.dw, ev.mouse.display);
			break;
		case ALLEGRO_EVENT_MOUSE_BUTTON_DOWN:
			OnMouseButtonDown(ev.mouse.x, ev.mouse.y, ev.mouse.z, ev.mouse.w, ev.mouse.button, ev.mouse.display);
			break;
		case ALLEGRO_EVENT_MOUSE_BUTTON_UP:
			OnMouseButtonUp(ev.mouse.x, ev.mouse.y, ev.mouse.z, ev.mouse.w, ev.mouse.button, ev.mouse.display);
			break;
		case ALLEGRO_EVENT_MOUSE_WARPED:
			OnMouseWarped(ev.mouse.x, ev.mouse.y, ev.mouse.z, ev.mouse.w, ev.mouse.dx, ev.mouse.dy, ev.mouse.dz, ev.mouse.dw, ev.mouse.display);
			break;
		case ALLEGRO_EVENT_MOUSE_ENTER_DISPLAY:
			OnMouseEnterDisplay(ev.mouse.x, ev.mouse.y, ev.mouse.z, ev.mouse.w, ev.mouse.display);
			break;
		case ALLEGRO_EVENT_MOUSE_LEAVE_DISPLAY:
			OnMouseLeaveDisplay(ev.mouse.x, ev.mouse.y, ev.mouse.z, ev.mouse.w, ev.mouse.display);
			break;
		case ALLEGRO_EVENT_TIMER:
			OnEventTimer(ev.timer.source, ev.timer.count);
			break;
		case ALLEGRO_EVENT_DISPLAY_EXPOSE:
			OnDisplayExpose(ev.display.source, ev.display.x, ev.display.y, ev.display.width, ev.display.height);
			break;
		case ALLEGRO_EVENT_DISPLAY_RESIZE:
			OnDisplayResize(ev.display.source, ev.display.x, ev.display.y, ev.display.width, ev.display.height);
			break;
		case ALLEGRO_EVENT_DISPLAY_CLOSE:
			OnDisplayClose(ev.display.source);
			Stop();
			break;
		case ALLEGRO_EVENT_DISPLAY_LOST:
			OnDisplayLost(ev.display.source);
			break;
		case ALLEGRO_EVENT_DISPLAY_FOUND:
			OnDisplayFound(ev.display.source);
			break;
		case ALLEGRO_EVENT_DISPLAY_SWITCH_OUT:
			OnDisplaySwitchOut(ev.display.source);
			break;
		case ALLEGRO_EVENT_DISPLAY_SWITCH_IN:
			OnDisplaySwitchIn(ev.display.source);
			break;
		case ALLEGRO_EVENT_DISPLAY_ORIENTATION:
			OnDisplayWindowOrientation(ev.display.source, ev.display.orientation);
			break;
		}
	}

	return isRunning;
}


// private AllegroGameEngine methods


void AllegroGameEngine::log(const string &message)
{
	cerr << message << endl;
}


// Event Handlers


void AllegroGameEngine::OnKeyDown(int keycode)
{
	keyboard.keycodes[keycode] = true;
	WidgetSwitch::OnKeyDown(keycode);
}


void AllegroGameEngine::OnKeyUp(int keycode)
{
	keyboard.keycodes[keycode] = false;
	WidgetSwitch::OnKeyUp(keycode);
}


void AllegroGameEngine::OnKeyChar(int keycode, int unicode, unsigned int modifiers, bool repeat)
{
	//keyboard.buffer.push_back(unicode);
	keyboard.modifiers = modifiers;
	WidgetSwitch::OnKeyChar(unicode);
}


ALLEGRO_FONT *AllegroGameEngine::getFont(const string &name)
{
	map<string, ALLEGRO_FONT*>::iterator it = fonts.find(name);
	if (it == fonts.end())
		return NULL;
	return it->second;
}


void AllegroGameEngine::OnJoystickAxis(ALLEGRO_JOYSTICK *id, int stick, int axis, float pos)
{
	joysticks[id].sticks[stick][axis] = pos;
	WidgetSwitch::OnJoystickAxis(axis, pos);
}


void AllegroGameEngine::OnJoystickButtonDown(ALLEGRO_JOYSTICK *id, int button)
{
	joysticks[id].buttons[button] = true;
	WidgetSwitch::OnJoystickButtonDown(button);
}


void AllegroGameEngine::OnJoystickButtonUp(ALLEGRO_JOYSTICK *id, int button)
{
	joysticks[id].buttons[button] = false;
	WidgetSwitch::OnJoystickButtonUp(button);
}


void AllegroGameEngine::OnJoystickConfiguration()
{
	al_reconfigure_joysticks();
}


void AllegroGameEngine::OnKeyDown(int keycode, ALLEGRO_DISPLAY *display)
{
	keyboard.keycodes[keycode] = true;
	WidgetSwitch::OnKeyDown(keycode);
	if (keycode == ALLEGRO_KEY_ESCAPE)
		Stop();
}


void AllegroGameEngine::OnKeyUp(int keycode, ALLEGRO_DISPLAY *display)
{
	keyboard.keycodes[keycode] = false;
	WidgetSwitch::OnKeyUp(keycode);
}


void AllegroGameEngine::OnKeyChar(int keycode, int unichar, int modifiers, bool repeat, ALLEGRO_DISPLAY *display)
{
	keyboard.buffer.push_back(unichar);
	WidgetSwitch::OnKeyChar(keycode);
}


void AllegroGameEngine::OnMouseAxes(int x, int y, int z, int w, int dx, int dy, int dz, int dw, ALLEGRO_DISPLAY *display)
{
	mouse.x = x;
	mouse.y = y;
	mouse.z = z;
	mouse.w = w;
	mouse.dx = dx;
	mouse.dy = dy;
	mouse.dz = dz;
	mouse.dw = dw;
	OnMouseMotion(x, y, dx, dy);
}


void AllegroGameEngine::OnMouseButtonDown(int x, int y, int z, int w, unsigned int button, ALLEGRO_DISPLAY *display)
{
	mouse.x = x;
	mouse.y = y;
	mouse.z = z;
	mouse.w = w;
	mouse.buttons[button] = true;
	WidgetSwitch::OnMouseButtonDown(x, y, button);
}


void AllegroGameEngine::OnMouseButtonUp(int x, int y, int z, int w, unsigned int button, ALLEGRO_DISPLAY *display)
{
	mouse.x = y;
	mouse.y = y;
	mouse.z = z;
	mouse.w = w;
	mouse.buttons[button] = false;
	WidgetSwitch::OnMouseButtonUp(x, y, button);
}


void AllegroGameEngine::OnMouseWarped(int x, int y, int z, int w, int dx, int dy, int dz, int dw, ALLEGRO_DISPLAY *display)
{
	OnMouseAxes(x, y, z, w, dx, dy, dz, dw, display);
}


void AllegroGameEngine::OnMouseEnterDisplay(int x, int y, int z, int w, ALLEGRO_DISPLAY *display)
{

}


void AllegroGameEngine::OnMouseLeaveDisplay(int x, int y, int z, int w, ALLEGRO_DISPLAY *display)
{

}


void AllegroGameEngine::OnEventTimer(ALLEGRO_TIMER *timer, int64_t count)
{

}


void AllegroGameEngine::OnDisplayExpose(ALLEGRO_DISPLAY *source, int x, int y, int width, int height)
{

}


void AllegroGameEngine::OnDisplayResize(ALLEGRO_DISPLAY *source, int x, int y, int width, int height)
{
	WidgetSwitch::OnDisplayResize(width, height);
}


void AllegroGameEngine::OnDisplayClose(ALLEGRO_DISPLAY *source)
{
	OnStop();
}


void AllegroGameEngine::OnDisplayLost(ALLEGRO_DISPLAY *source)
{

}


void AllegroGameEngine::OnDisplayFound(ALLEGRO_DISPLAY *source)
{

}


void AllegroGameEngine::OnDisplaySwitchOut(ALLEGRO_DISPLAY *source)
{

}


void AllegroGameEngine::OnDisplaySwitchIn(ALLEGRO_DISPLAY *source)
{

}


void AllegroGameEngine::OnDisplayWindowOrientation(ALLEGRO_DISPLAY *source, int orientation)
{

}
