#ifndef XOR_ALLEGRO_H
#define XOR_ALLEGRO_H


#include "stdafx.h"
#include <allegro5/allegro.h>
#include <allegro5/allegro_font.h>
#include <allegro5/allegro_ttf.h>
#include <allegro5/allegro_image.h>
#include <allegro5/allegro_audio.h>
#include <allegro5/allegro_acodec.h>
#include <allegro5/allegro_primitives.h>


#include "xor_widget.h"
#include "xor_input_devices.h"


class AllegroGameEngine : public WidgetSwitch
{
private:
	double t0 = 0.0;
	double t1 = 0.0;
	double dt = 0.0;
	double elapsedTime = 0.0;

	ALLEGRO_DISPLAY *display = NULL;
	ALLEGRO_EVENT_QUEUE *eventQueue = NULL;
	
	map<string, ALLEGRO_FONT *> fonts;

	map<ALLEGRO_JOYSTICK *, Joystick> joysticks;
	Mouse mouse;
	Keyboard keyboard;

	bool isInitted = false;
	bool isRunning = false;

	int screenWidth;
	int screenHeight;

public:
	AllegroGameEngine();
	~AllegroGameEngine();

	void log(const string &message);
	ALLEGRO_EVENT_QUEUE *getEventQueue() { return eventQueue; }

	int Init(int width, int height);
	void Kill();
	void Start();
	void Stop();

	bool MainLoopIteration();

	int getScreenWidth() const { return screenWidth; }
	int getScreenHeight() const { return screenHeight; }

	// Widget Event Handlers
	void OnKeyDown(int keycode);
	void OnKeyUp(int keycode);
	void OnKeyChar(int keycode, int unicode, unsigned int modifiers, bool repeat);

	// Allegro Event Handlers
	virtual void OnJoystickAxis(ALLEGRO_JOYSTICK *id, int stick, int axis, float pos);
	virtual void OnJoystickButtonDown(ALLEGRO_JOYSTICK *id, int button);
	virtual void OnJoystickButtonUp(ALLEGRO_JOYSTICK *id, int button);
	virtual void OnJoystickConfiguration();
	virtual void OnKeyDown(int keycode, ALLEGRO_DISPLAY *display);
	virtual void OnKeyUp(int keycode, ALLEGRO_DISPLAY *display);
	virtual void OnKeyChar(int keycode, int unichar, int modifiers, bool repeat, ALLEGRO_DISPLAY *display);
	virtual void OnMouseAxes(int x, int y, int z, int w, int dx, int dy, int dz, int dw, ALLEGRO_DISPLAY *display);
	virtual void OnMouseButtonDown(int x, int y, int z, int w, unsigned int button, ALLEGRO_DISPLAY *display);
	virtual void OnMouseButtonUp(int x, int y, int z, int w, unsigned int button, ALLEGRO_DISPLAY *display);
	virtual void OnMouseWarped(int x, int y, int z, int w, int dx, int dy, int dz, int dw, ALLEGRO_DISPLAY *display);
	virtual void OnMouseEnterDisplay(int x, int y, int z, int w, ALLEGRO_DISPLAY *display);
	virtual void OnMouseLeaveDisplay(int x, int y, int z, int w, ALLEGRO_DISPLAY *display);
	virtual void OnEventTimer(ALLEGRO_TIMER *timer, int64_t count);
	virtual void OnDisplayExpose(ALLEGRO_DISPLAY *source, int x, int y, int width, int height);
	virtual void OnDisplayResize(ALLEGRO_DISPLAY *source, int x, int y, int width, int height);
	virtual void OnDisplayClose(ALLEGRO_DISPLAY *source);
	virtual void OnDisplayLost(ALLEGRO_DISPLAY *source);
	virtual void OnDisplayFound(ALLEGRO_DISPLAY *source);
	virtual void OnDisplaySwitchOut(ALLEGRO_DISPLAY *source);
	virtual void OnDisplaySwitchIn(ALLEGRO_DISPLAY *source);
	virtual void OnDisplayWindowOrientation(ALLEGRO_DISPLAY *source, int orientation);

	void ResetTimer() { elapsedTime = 0.0; }
	double GetElapsedTime() const { return elapsedTime; }
	double GetDeltaTime() const { return dt; }

	ALLEGRO_FONT *getFont(const string &name);
};


extern AllegroGameEngine age;


#endif
