#include "stdafx.h"
#include "xor_sound.h"


SoundSystem SS;


SoundSystem::SoundSystem()
{

}


SoundSystem::~SoundSystem()
{

}


void SoundSystem::Init()
{
	al_init_acodec_addon();
	
	if (!al_install_audio())
	{
		age.log("SoundSystem::Init() --> Unable to initialize audio");
		return;
	}

	al_reserve_samples(16);

	voice = al_create_voice(44100, ALLEGRO_AUDIO_DEPTH_INT16, ALLEGRO_CHANNEL_CONF_2);
	if (!voice)
	{
		age.log("SoundSystem::Init() --> Unable to create voice");
		Kill();
		return;
	}

	//mixer = al_create_mixer(44100, ALLEGRO_AUDIO_DEPTH_FLOAT32, ALLEGRO_CHANNEL_CONF_2);
	//if (!mixer)
	//{
	//	age.log("SoundSystem::Init() --> Unable to create mixer");
	//	Kill();
	//	return;
	//}

	//al_attach_mixer_to_voice(mixer, voice);
	//al_set_default_mixer(mixer);
}


void SoundSystem::Kill()
{
	StopMusic();
	StopSamples();

	if (voice)
	{
		al_destroy_voice(voice);
		voice = NULL;
	}

	if (mixer)
	{
		al_destroy_mixer(mixer);
		mixer = NULL;
	}

	al_uninstall_audio();
}


ALLEGRO_SAMPLE *SoundSystem::FindSample(const string &name)
{
	map<string, ALLEGRO_SAMPLE*>::iterator it = samples.find(name);
	if (it == samples.end())
		return NULL;
	return it->second;
}


void SoundSystem::DeleteSample(const string &name)
{
	ALLEGRO_SAMPLE *sample = FindSample(name);
	if (!sample)
		return;
	al_destroy_sample(sample);
	samples[name] = NULL;
}


void SoundSystem::LoadSample(const string &name, const string &filename)
{
	DeleteSample(name);
	ALLEGRO_SAMPLE *sample = al_load_sample(filename.c_str());
	if (!sample)
	{
		return;
	}
	samples[name] = sample;
}


void SoundSystem::PlaySample(const string &name, float gain, float pan, float speed, bool loop)
{
	ALLEGRO_SAMPLE_ID sampleID;
	ALLEGRO_SAMPLE *sample = FindSample(name);
	if (!sample)
	{
		return;
	}
	al_play_sample(sample, gain, pan, speed, loop ? ALLEGRO_PLAYMODE_LOOP : ALLEGRO_PLAYMODE_ONCE, &sampleID);
	sampleIDs.push_back(sampleID);
}


void SoundSystem::StopSamples()
{
	al_stop_samples();
	for_each(sampleIDs.begin(), sampleIDs.end(), [](ALLEGRO_SAMPLE_ID &sampleID){
		al_stop_sample(&sampleID);
	});
	sampleIDs.clear();
}


void SoundSystem::PlayMusic(const string &filename)
{
	StopMusic();
	audioStream = al_load_audio_stream(filename.c_str(), 4, 2048);
	if (!audioStream)
		return;

	al_attach_audio_stream_to_mixer(audioStream, al_get_default_mixer());

	al_register_event_source(age.getEventQueue(), al_get_audio_stream_event_source(audioStream));
}


void SoundSystem::StopMusic()
{
	if (audioStream)
	{
		al_detach_audio_stream(audioStream);
		al_destroy_audio_stream(audioStream);
		audioStream = NULL;
	}
}
